"""Agents management commands - lazy loading subcommands."""

from typing import Optional, get_args, Annotated, Literal
import typer
from machineconfig.scripts.python.helpers.helpers_agents.fire_agents_helper_types import AGENTS, HOST, PROVIDER


def agents_create(
    agent: Annotated[AGENTS, typer.Option(..., "--agents", "-a", help=f"Agent type. One of {', '.join(get_args(AGENTS)[:3])}")],
    host: Annotated[HOST, typer.Option(..., "--host", "-h", help=f"Machine to run agents on. One of {', '.join(get_args(HOST))}")],
    model: Annotated[str, typer.Option(..., "--model", "-m", help="Model to use (for crush agent).")],
    provider: Annotated[PROVIDER, typer.Option(..., "--provider", "-p", help=f"Provider to use (for crush agent). One of {', '.join(get_args(PROVIDER)[:3])}")],
    context_path: Annotated[Optional[str], typer.Option(..., "--context-path", "-c", help="Path to the context file/folder, defaults to .ai/todo/")] = None,
    separator: Annotated[str, typer.Option(..., "--separator", "-s", help="Separator for context")] = "\n",
    agent_load: Annotated[int, typer.Option(..., "--agent-load", "-al", help="Number of tasks per prompt")] = 13,
    prompt: Annotated[Optional[str], typer.Option(..., "--prompt", "-P", help="Prompt prefix as string")] = None,
    prompt_path: Annotated[Optional[str], typer.Option(..., "--prompt-path", "-pp", help="Path to prompt file")] = None,
    job_name: Annotated[str, typer.Option(..., "--job-name", "-j", help="Job name")] = "AI_Agents",
    separate: Annotated[bool, typer.Option(..., "--separate", "-S", help="Keep prompt material in separate file to the context.")] = True,
    output_path: Annotated[Optional[str], typer.Option(..., "--output-path", "-o", help="Path to write the layout.json file")] = None,
    agents_dir: Annotated[Optional[str], typer.Option(..., "--agents-dir", "-ad", help="Directory to store agent files. If not provided, will be constructed automatically.")] = None,
) -> None:
    """Create agents layout file, ready to run."""
    from machineconfig.scripts.python.helpers.helpers_agents.agents_impl import agents_create as impl
    try:
        impl(agent=agent, host=host, model=model, provider=provider, context_path=context_path, separator=separator, agent_load=agent_load, prompt=prompt, prompt_path=prompt_path, job_name=job_name, separate=separate, output_path=output_path, agents_dir=agents_dir)
    except ValueError as e:
        raise typer.BadParameter(str(e)) from e
    except RuntimeError as e:
        typer.echo(str(e))
        raise typer.Exit(1) from e
    
def collect(
    agent_dir: Annotated[str, typer.Argument(..., help="Path to the agent directory containing the prompts folder")],
    output_path: Annotated[str, typer.Argument(..., help="Path to write the concatenated material files")],
    separator: Annotated[str, typer.Option(..., help="Separator to use when concatenating material files")] = "\n",
) -> None:
    """Collect all material files from an agent directory and concatenate them."""
    from machineconfig.scripts.python.helpers.helpers_agents.agents_impl import collect as impl
    try:
        impl(agent_dir=agent_dir, output_path=output_path, separator=separator)
    except ValueError as e:
        raise typer.BadParameter(str(e)) from e


def make_agents_command_template() -> None:
    """Create a template for fire agents."""
    from machineconfig.scripts.python.helpers.helpers_agents.agents_impl import make_agents_command_template as impl
    try:
        impl()
    except ValueError as e:
        raise typer.BadParameter(str(e)) from e
    except RuntimeError as e:
        typer.echo(str(e))
        raise typer.Exit(1) from e


def init_config(root: Annotated[Optional[str], typer.Option(..., "--root", "-r", help="Root directory of the repository to initialize AI configs in. Defaults to current directory.")] = None) -> None:
    """Initialize AI configurations in the current repository."""
    from machineconfig.scripts.python.helpers.helpers_agents.agents_impl import init_config as impl
    impl(root=root)


def make_todo_files(
    pattern: Annotated[str, typer.Argument(help="Pattern or keyword to match files by")] = ".py",
    repo: Annotated[str, typer.Argument(help="Repository path. Can be any directory within a git repository.")] = ".",
    strategy: Annotated[Literal["name", "keywords"], typer.Option("-s", "--strategy", help="Strategy to filter files: 'name' for filename matching, 'keywords' for content matching")] = "name",
    exclude_init: Annotated[bool, typer.Option("-x", "--exclude-init", help="Exclude __init__.py files from the checklist")] = True,
    include_line_count: Annotated[bool, typer.Option("-l", "--line-count", help="Include line count column in the output")] = False,
    output_path: Annotated[str, typer.Option("-o", "--output-path", help="Base path for output files relative to repo root")] = ".ai/todo/files",
    format_type: Annotated[Literal["csv", "md", "txt"], typer.Option("-f", "--format", help="Output format: csv, md (markdown), or txt")] = "md",
    split_every: Annotated[Optional[int], typer.Option("--split-every", "-e", help="Split output into multiple files, each containing at most this many results")] = None,
    split_to: Annotated[Optional[int], typer.Option("--split-to", "-t", help="Split output into exactly this many files")] = None,
) -> None:
    """Generate checklist with Python and shell script files in the repository filtered by pattern."""
    from machineconfig.scripts.python.ai.utils.generate_files import make_todo_files as impl
    impl(pattern=pattern, repo=repo, strategy=strategy, exclude_init=exclude_init, include_line_count=include_line_count, output_path=output_path, format_type=format_type, split_every=split_every, split_to=split_to)


def create_symlink_command(
    num: Annotated[int, typer.Argument(help="Number of symlinks to create (1-5).")] = 5,
) -> None:
    """Create symlinks to repo_root at ~/code_copies/${repo_name}_copy_{i}."""
    from machineconfig.scripts.python.ai.utils.generate_files import create_symlink_command as impl
    impl(num=num)


def get_app() -> typer.Typer:
    agents_app = typer.Typer(help="🤖 AI Agents management subcommands", no_args_is_help=True, add_help_option=True, add_completion=False)
    sep = "\n"
    agents_full_help = f"""
[c] Create agents layout file, ready to run.
{sep}
PROVIDER options: {', '.join(get_args(PROVIDER))}
{sep}
AGENT options: {', '.join(get_args(AGENTS))}
"""
    agents_create.__doc__ = agents_full_help
    agents_app.command("create", no_args_is_help=True, help=agents_create.__doc__, short_help="[c] Create agents layout file, ready to run.")(agents_create)
    agents_app.command("c", no_args_is_help=True, help=agents_create.__doc__, hidden=True)(agents_create)
    agents_app.command("collect", no_args_is_help=True, help=collect.__doc__, short_help="[T] Collect all agent materials into a single file.")(collect)
    agents_app.command("T", no_args_is_help=True, help=collect.__doc__, hidden=True)(collect)
    agents_app.command("make-template", no_args_is_help=False, help=make_agents_command_template.__doc__, short_help="[t] Create a template for fire agents")(make_agents_command_template)
    agents_app.command("t", no_args_is_help=False, help=make_agents_command_template.__doc__, hidden=True)(make_agents_command_template)
    agents_app.command("make-config", no_args_is_help=False, help=init_config.__doc__, short_help="[g] Initialize AI configurations in the current repository")(init_config)
    agents_app.command("g", no_args_is_help=False, help=init_config.__doc__, hidden=True)(init_config)
    agents_app.command("make-todo", no_args_is_help=True, short_help="[d] Generate a markdown file listing all Python files in the repo")(make_todo_files)
    agents_app.command("d", no_args_is_help=True, hidden=True)(make_todo_files)
    agents_app.command(name="make-symlinks", no_args_is_help=True, short_help="[s] Create symlinks to the current repo in ~/code_copies/")(create_symlink_command)
    agents_app.command(name="s", no_args_is_help=True, hidden=True)(create_symlink_command)
    return agents_app


def main() -> None:
    agents_app = get_app()
    agents_app()


if __name__ == "__main__":
    pass




import typer
from typing import Annotated, Literal


def share_file_receive(code_args: Annotated[list[str], typer.Argument(help="Receive code or relay command. Examples: '7121-donor-olympic-bicycle' or '--relay 10.17.62.206:443 7121-donor-olympic-bicycle'")],
) -> None:
    """Receive a file using croc with relay server.
Usage examples:
    devops network receive 7121-donor-olympic-bicycle
    devops network receive -- --relay 10.17.62.206:443 7121-donor-olympic-bicycle
    devops network receive -- croc --relay 10.17.62.206:443 7121-donor-olympic-bicycle
"""
    from machineconfig.utils.installer_utils.installer_cli import install_if_missing
    install_if_missing(which="croc")
    import platform
    import sys

    is_windows = platform.system() == "Windows"

    # If no args passed via typer, try to get them from sys.argv directly
    # This handles the case where -- was used and arguments weren't parsed by typer
    if not code_args or (len(code_args) == 1 and code_args[0] in ['--relay', 'croc']):
        # Find the index of 'rx' or 'receive' in sys.argv and get everything after it
        try:
            for i, arg in enumerate(sys.argv):
                if arg in ['rx', 'receive', 'r'] and i + 1 < len(sys.argv):
                    code_args = sys.argv[i + 1:]
                    break
        except Exception:
            pass

    # Join all arguments
    input_str = " ".join(code_args)
    tokens = input_str.split()

    # Parse input to extract relay server and secret code
    relay_server: str | None = None
    secret_code: str | None = None

    # Remove 'croc' and 'export' from tokens if present
    tokens = [t for t in tokens if t not in ['croc', 'export']]

    # Look for --relay flag and capture next token
    relay_idx = -1
    for i, token in enumerate(tokens):
        if token == '--relay' and i + 1 < len(tokens):
            relay_server = tokens[i + 1]
            relay_idx = i
            break

    # Look for CROC_SECRET= prefix in any token
    for token in tokens:
        if token.startswith('CROC_SECRET='):
            secret_code = token.split('=', 1)[1].strip('"').strip("'")
            break

    # If no secret code found yet, look for tokens with dashes (typical pattern: number-word-word-word)
    # Skip relay server and relay flag
    if not secret_code:
        for i, token in enumerate(tokens):
            if '-' in token and not token.startswith('-') and token != relay_server:
                if relay_idx >= 0 and (i == relay_idx or i == relay_idx + 1):
                    continue  # Skip relay server parts
                secret_code = token
                break

    if not secret_code and not relay_server:
        typer.echo(f"❌ Error: Could not parse croc receive code from input: {input_str}", err=True)
        typer.echo("Usage:", err=True)
        typer.echo("  devops network receive 7121-donor-olympic-bicycle", err=True)
        typer.echo("  devops network receive -- --relay 10.17.62.206:443 7121-donor-olympic-bicycle", err=True)
        raise typer.Exit(code=1)

    # Build the appropriate script for current OS
    if is_windows:
        # Windows PowerShell format: croc --relay server:port secret-code --yes
        relay_arg = f"--relay {relay_server}" if relay_server else ""
        code_arg = f"{secret_code}" if secret_code else ""
        script = f"""croc {relay_arg} {code_arg} --yes""".strip()
    else:
        # Linux/macOS Bash format: CROC_SECRET="secret-code" croc --relay server:port --yes
        relay_arg = f"--relay {relay_server}" if relay_server else ""
        if secret_code:
            script = f"""export CROC_SECRET="{secret_code}"
croc {relay_arg} --yes""".strip()
        else:
            script = f"""croc {relay_arg} --yes""".strip()

    from machineconfig.utils.code import exit_then_run_shell_script, print_code
    print_code(code=script, desc="🚀 Receiving file with croc", lexer="bash" if platform.system() != "Windows" else "powershell")
    exit_then_run_shell_script(script=script, strict=False)


def share_file_send(path: Annotated[str, typer.Argument(help="Path to the file or directory to send")],
                    zip_folder: Annotated[bool, typer.Option("--zip", help="Zip folder before sending")] = False,
                    code: Annotated[str | None, typer.Option("--code", "-c", help="Codephrase used to connect to relay")] = None,
                    text: Annotated[str | None, typer.Option("--text", "-t", help="Send some text")] = None,
                    qrcode: Annotated[bool, typer.Option("--qrcode", "--qr", help="Show receive code as a qrcode")] = False,
                    backend: Annotated[Literal["wormhole", "w", "croc", "c"], typer.Option("--backend", "-b", help="Backend to use")] = "croc",
                    ) -> None:
    """Send a file using croc with relay server."""
    import platform

    match backend:
        case "wormhole" | "w":
            if code is None: code_line = ""
            else: code_line = f"--code {code}"
            if text is not None: text_line = f"--text '{text}'"
            else: text_line = f"'{path}'"
            script = f"""
uvx magic-wormhole send {code_line} {text_line}
"""
            print(f"🚀 Sending file: {path}. Use: uvx magic-wormhole receive ")
        case "croc" | "c":
            from machineconfig.utils.installer_utils.installer_cli import install_if_missing
            install_if_missing(which="croc")

            # Get relay server IP from environment or use default
            import machineconfig.scripts.python.helpers.helpers_network.address as helper
            res = helper.select_lan_ipv4(prefer_vpn=False)
            if res is None:
                typer.echo("❌ Error: Could not determine local LAN IPv4 address for relay.", err=True)
                raise typer.Exit(code=1)
            local_ip_v4 = res
            relay_port = "443"
            is_windows = platform.system() == "Windows"
            # Build command parts
            relay_arg = f"--relay {local_ip_v4}:{relay_port} --ip {local_ip_v4}:{relay_port}"
            zip_arg = "--zip" if zip_folder else ""
            text_arg = f"--text '{text}'" if text else ""
            qrcode_arg = "--qrcode" if qrcode else ""
            path_arg = f"{path}" if not text else ""

            if is_windows:
                # Windows PowerShell format
                code_arg = f"--code {code}" if code else ""
                script = f"""croc {relay_arg} send {zip_arg} {code_arg} {qrcode_arg} {text_arg} {path_arg}"""
            else:
                # Linux/macOS Bash format
                if code:
                    script = f"""
export CROC_SECRET="{code}"
croc {relay_arg} send {zip_arg} {qrcode_arg} {text_arg} {path_arg}"""
                else:
                    script = f"""croc {relay_arg} send {zip_arg} {qrcode_arg} {text_arg} {path_arg}"""

    typer.echo(f"🚀 Sending file: {path}. Use: devops network receive")
    from machineconfig.utils.code import exit_then_run_shell_script, print_code
    print_code(code=script, desc="🚀 sending file with croc", lexer="bash" if platform.system() != "Windows" else "powershell")
    exit_then_run_shell_script(script=script, strict=False)



import os
import platform
import stat
import shutil
import subprocess
from pathlib import Path, PureWindowsPath


def ensure_relative_path(requested: Path | str) -> Path:
    path = Path(requested)
    if path.is_absolute():
        raise ValueError("paths must be relative to the home directory")
    if any(part == ".." for part in path.parts):
        raise ValueError("paths must stay within the home directory")
    return path


def remove_path(path: Path) -> None:
    if path.is_symlink() or path.is_file():
        path.unlink()
        return
    shutil.rmtree(path)


def ensure_wsl_environment() -> None:
    if os.environ.get("WSL_DISTRO_NAME"):
        return
    if "microsoft" in platform.release().lower():
        return
    raise RuntimeError("copy_when_inside_wsl must run inside WSL")


def ensure_windows_environment() -> None:
    if os.name != "nt":
        raise RuntimeError("copy_when_inside_windows must run inside Windows")
    if os.environ.get("WSL_DISTRO_NAME"):
        raise RuntimeError("copy_when_inside_windows must run inside Windows")


def ensure_linux_environment() -> None:
    if platform.system() != "Linux":
        raise RuntimeError("change_ssh_port requires Linux environment")


def infer_windows_home_from_permissions(windows_username: str | None) -> Path:
    base_dir = Path("/mnt/c/Users")
    try:
        entries = list(base_dir.iterdir())
    except FileNotFoundError as exc:
        raise RuntimeError("unable to find /mnt/c/Users") from exc
    if windows_username:
        candidate = base_dir / windows_username
        if candidate.is_dir():
            return candidate
        raise RuntimeError(f"specified Windows user directory not found: {candidate}")
    candidates: list[Path] = []
    for entry in entries:
        if not entry.is_dir():
            continue
        if entry.name.lower() == "public" or entry.name.lower() == "all users":
            continue
        try:
            mode = stat.S_IMODE(entry.stat().st_mode)
        except OSError:
            continue
        if mode == 0o777:
            candidates.append(entry)
    if len(candidates) != 1:
        wsl_user = os.environ.get("USER") or os.environ.get("LOGNAME")
        if wsl_user:
            for candidate in candidates:
                if candidate.name == wsl_user:
                    return candidate
        non_default = [c for c in candidates if c.name.lower() not in ("default", "default user")]
        if len(non_default) == 1:
            return non_default[0]
        options = ", ".join(sorted(candidate.name for candidate in candidates)) or "none"
        raise RuntimeError(f"unable to infer Windows home directory (candidates: {options})")
    return candidates[0]


def resolve_windows_home_from_wsl(windows_username: str | None) -> Path:
    if windows_username:
        return infer_windows_home_from_permissions(windows_username)
    user_profile = os.environ.get("USERPROFILE")
    if user_profile:
        windows_path = PureWindowsPath(user_profile)
        drive = windows_path.drive
        if drive:
            drive_letter = drive.rstrip(":").lower()
            tail = Path(*windows_path.parts[1:])
            candidate = Path("/mnt") / drive_letter / tail
            if candidate.exists():
                return candidate
    return infer_windows_home_from_permissions(windows_username)


def decode_wsl_output(raw_bytes: bytes) -> str:
    try:
        return raw_bytes.decode("utf-16-le")
    except UnicodeDecodeError:
        return raw_bytes.decode()


def get_single_wsl_distribution() -> str:
    process = subprocess.run(["wsl.exe", "-l"], capture_output=True, text=False, check=True)
    stdout = decode_wsl_output(process.stdout).replace("\ufeff", "")
    distributions: list[str] = []
    for raw_line in stdout.splitlines():
        line = raw_line.strip()
        if not line or line.lower().startswith("windows subsystem for linux"):
            continue
        normalized = line.lstrip("* ").replace("(Default)", "").strip()
        if normalized:
            distributions.append(normalized)
    if len(distributions) != 1:
        raise RuntimeError("unable to pick a single WSL distribution")
    return distributions[0]


def resolve_wsl_home_on_windows() -> Path:
    distribution = get_single_wsl_distribution()
    home_root = Path(rf"\\wsl$\{distribution}\home")
    try:
        entries = list(home_root.iterdir())
    except FileNotFoundError as exc:
        raise RuntimeError(f"unable to locate WSL home directories for {distribution}") from exc
    except OSError as exc:
        raise RuntimeError(f"unable to inspect WSL home directories for {distribution}") from exc
    user_dirs = [entry for entry in entries if entry.is_dir()]
    if len(user_dirs) != 1:
        options = ", ".join(sorted(entry.name for entry in user_dirs)) or "none"
        raise RuntimeError(f"unable to infer WSL user directory (candidates: {options})")
    return user_dirs[0]


def quote_for_powershell(path: Path) -> str:
    return "'" + str(path).replace("'", "''") + "'"


def run_windows_copy_command(source_path: Path, target_path: Path) -> None:
    source_is_dir = source_path.is_dir()
    parent_literal = quote_for_powershell(target_path.parent)
    source_literal = quote_for_powershell(source_path)
    target_literal = quote_for_powershell(target_path)
    script = (
        "$ErrorActionPreference = 'Stop'; "
        f"New-Item -ItemType Directory -Path {parent_literal} -Force | Out-Null; "
        f"Copy-Item -LiteralPath {source_literal} -Destination {target_literal}"
        f"{' -Recurse' if source_is_dir else ''} -Force"
    )
    print(f"Copying {source_path} to {target_path}")
    subprocess.run(["powershell.exe", "-NoLogo", "-NoProfile", "-Command", script], check=True)


def ensure_symlink(link_path: Path, target_path: Path) -> bool:
    if not target_path.exists():
        raise FileNotFoundError(target_path)
    if link_path.is_symlink():
        existing_target = Path(os.path.realpath(link_path))
        desired_target = Path(os.path.realpath(target_path))
        if os.path.normcase(str(existing_target)) == os.path.normcase(str(desired_target)):
            return False
        link_path.unlink()
    elif link_path.exists():
        raise FileExistsError(link_path)
    link_path.symlink_to(target_path, target_is_directory=True)
    return True


def parse_port_spec(port_spec: str) -> list[int]:
    ports: list[int] = []
    for part in port_spec.split(","):
        part = part.strip()
        if not part:
            continue
        if "-" in part:
            range_parts = part.split("-", maxsplit=1)
            start = int(range_parts[0].strip())
            end = int(range_parts[1].strip())
            if start > end:
                raise ValueError(f"Invalid port range: {part} (start > end)")
            ports.extend(range(start, end + 1))
        else:
            ports.append(int(part))
    return ports


def normalize_port_spec_for_firewall(port_spec: str) -> tuple[str, str]:
    """Validates and normalizes port spec for New-NetFirewallRule -LocalPort parameter.
    Returns (normalized_spec, human_readable_description).
    New-NetFirewallRule natively supports ranges (e.g. '5000-5020') and comma-separated values.
    """
    parts: list[str] = []
    for raw_part in port_spec.split(","):
        part = raw_part.strip()
        if not part:
            continue
        if "-" in part:
            range_parts = part.split("-", maxsplit=1)
            start = int(range_parts[0].strip())
            end = int(range_parts[1].strip())
            if start < 1 or start > 65535 or end < 1 or end > 65535:
                raise ValueError(f"Port numbers must be between 1 and 65535: {part}")
            if start > end:
                raise ValueError(f"Invalid port range: {part} (start > end)")
            parts.append(f"{start}-{end}")
        else:
            port = int(part)
            if port < 1 or port > 65535:
                raise ValueError(f"Invalid port number: {port}")
            parts.append(str(port))
    if not parts:
        raise ValueError("No valid ports provided")
    normalized = ",".join(parts)
    description = normalized.replace(",", ", ")
    return normalized, description



"""
Report Utilities
================

This module provides functionality to generate reports for installed applications.
"""

from pathlib import Path
from typing import Optional, TypedDict
from rich.console import Console
from rich.panel import Panel

console = Console()

class AppData(TypedDict):
    app_name: str
    version: Optional[str]
    positive_pct: Optional[float]
    scan_time: str
    app_path: str
    app_url: str

def generate_markdown_report(data: list[AppData], output_path: Path) -> None:
    """Generates a Markdown table from the app data."""
    if not data:
        return

    keys = list(data[0].keys())
    header = "| " + " | ".join(key.replace("_", " ").title() for key in keys) + " |"
    separator = "| " + " | ".join("---" for _ in keys) + " |"
    rows = []
    for row in data:
        # Fix: row is a dict, keys are strings.
        row_values = [str(row.get(key, '')) for key in keys]
        rows.append("| " + " | ".join(row_values) + " |")
    
    content = "\n".join([header, separator] + rows)
    output_path.write_text(content, encoding="utf-8")
    console.print(Panel(content, title="Safety Report Summary", expand=False))




import typer
from typing import Annotated, Literal


def copy_both_assets():
    import machineconfig.profile.create_helper as create_helper
    create_helper.copy_assets_to_machine(which="scripts")
    create_helper.copy_assets_to_machine(which="settings")


def init(which: Annotated[Literal["init", "ia", "live", "wrap"], typer.Argument(..., help="Comma-separated list of script names to run all initialization scripts.")] = "init",
         run: Annotated[bool, typer.Option("--run/--no-run", "-r/-nr", help="Run the script after displaying it.")] = False,
                        ) -> None:
    import platform
    if platform.system() == "Linux" or platform.system() == "Darwin":
        match which:
            case "init":
                import machineconfig.settings as module
                from pathlib import Path
                if platform.system() == "Darwin":
                    init_path = Path(module.__file__).parent.joinpath("shells", "zsh", "init.sh")
                else: init_path = Path(module.__file__).parent.joinpath("shells", "bash", "init.sh")
                script = init_path.read_text(encoding="utf-8")
            case "ia":
                from machineconfig.setup_linux import INTERACTIVE as script_path
                script = script_path.read_text(encoding="utf-8")
            case "live":
                from machineconfig.setup_linux import LIVE as script_path
                script = script_path.read_text(encoding="utf-8")
            case _:
                typer.echo("Unsupported shell script for Linux.")
                raise typer.Exit(code=1)

    elif platform.system() == "Windows":
        match which:
            case "init":
                import machineconfig.settings as module
                from pathlib import Path
                init_path = Path(module.__file__).parent.joinpath("shells", "powershell", "init.ps1")
                script = init_path.read_text(encoding="utf-8")
            case "ia":
                from machineconfig.setup_windows import INTERACTIVE as script_path
                script = script_path.read_text(encoding="utf-8")
            case "live":
                from machineconfig.setup_windows import LIVE as script_path
                script = script_path.read_text(encoding="utf-8")
            case _:
                typer.echo("Unsupported shell script for Windows.")
                raise typer.Exit(code=1)
    else:
        # raise NotImplementedError("Unsupported platform")
        typer.echo("Unsupported platform for init scripts.")
        raise typer.Exit(code=1)
    if run:
        from machineconfig.utils.code import exit_then_run_shell_script
        exit_then_run_shell_script(script, strict=True)
    else:
        print(script)


def update(copy_assets: Annotated[bool, typer.Option("--assets-copy/--no-assets-copy", "-a/-na", help="Copy (overwrite) assets to the machine after the update")] = True,
           link_public_configs: Annotated[bool, typer.Option("--link-public-configs/--no-link-public-configs", "-b/-nb", help="Link public configs after update (overwrites your configs!)")] = False,
           ):
    """🔄 UPDATE uv and machineconfig"""
    from pathlib import Path
    if Path.home().joinpath("code", "machineconfig").exists():
        shell_script = """
uv self update
cd ~/code/machineconfig
git pull
uv tool install --no-cache --upgrade --editable $HOME/code/machineconfig
    """
    else:
        shell_script = """
uv self update
uv tool install --no-cache --upgrade machineconfig
    """
    import platform
    if platform.system() == "Windows":
        from machineconfig.utils.code import exit_then_run_shell_script, get_uv_command_executing_python_script
        from machineconfig.utils.meta import lambda_to_python_script
        python_script = lambda_to_python_script(lambda: copy_both_assets(),
                                                in_global=True, import_module=False)
        uv_command, _py_file = get_uv_command_executing_python_script(python_script=python_script, uv_with=["machineconfig"], uv_project_dir=None)
        exit_then_run_shell_script(shell_script + "\n" + uv_command, strict=True)
    else:
        from machineconfig.utils.code import run_shell_script
        run_shell_script(shell_script)
        if copy_assets:
            copy_both_assets()
        if link_public_configs:
            import machineconfig.profile.create_links_export as create_links_export
            create_links_export.main_public_from_parser(method="copy", on_conflict="overwrite-default-path", which="all", interactive=False)


def install(copy_assets: Annotated[bool, typer.Option("--copy-assets/--no-assets-copy", "-a/-na", help="Copy (overwrite) assets to the machine after the update")] = True,
            dev: Annotated[bool, typer.Option("--dev", "-d", help="Install from local development code instead of PyPI")] = False,
    
            ):
    """📋 CLONE machienconfig locally and incorporate to shell profile for faster execution and nightly updates."""
    from machineconfig.utils.code import run_shell_script, get_uv_command
    from pathlib import Path
    import platform
    if dev and not Path.home().joinpath("code/machineconfig").exists():
        # clone: https://github.com/thisismygitrepo/machineconfig.git
        import git
        repo_parent = Path.home().joinpath("code")
        repo_parent.mkdir(parents=True, exist_ok=True)
        git.Repo.clone_from("https://github.com/thisismygitrepo/machineconfig.git", str(repo_parent.joinpath("machineconfig")))
    uv_command = get_uv_command(platform=platform.system())
    if Path.home().joinpath("code/machineconfig").exists():
        run_shell_script(f""" {uv_command} tool install --upgrade --editable "{str(Path.home().joinpath("code/machineconfig"))}" """)
    else:
        run_shell_script(rf""" {uv_command} tool install --upgrade "machineconfig>=8.28" """)
    if copy_assets:
        from machineconfig.profile.create_shell_profile import create_default_shell_profile
        create_default_shell_profile()   # involves copying assets too



def interactive():
    """🤖 INTERACTIVE configuration of machine."""
    from machineconfig.scripts.python.interactive import main
    main()

def status():
    """📊 STATUS of machine, shell profile, apps, symlinks, dotfiles, etc."""
    import machineconfig.scripts.python.helpers.helpers_devops.devops_status as helper
    helper.main()


def navigate():
    """📚 NAVIGATE command structure with TUI"""
    import machineconfig.scripts.python as navigator
    from pathlib import Path
    path = Path(navigator.__file__).resolve().parent.joinpath("devops_navigator.py")
    from machineconfig.utils.code import exit_then_run_shell_script
    if Path.home().joinpath("code/machineconfig").exists(): executable = f"""--project "{str(Path.home().joinpath("code/machineconfig"))}" --with textual"""
    else: executable = """--with "machineconfig>=8.28,textual" """
    exit_then_run_shell_script(f"""uv run {executable} {path}""")

def readme():
    from rich.console import Console
    from rich.markdown import Markdown
    import requests

    # URL of the raw README.md file
    url_readme = "https://raw.githubusercontent.com/thisismygitrepo/machineconfig/refs/heads/main/README.md"

    # Fetch the content
    response = requests.get(url_readme)
    response.raise_for_status()  # Raise an error for bad responses

    # Parse markdown
    md = Markdown(response.text)

    # Render in terminal
    console = Console()
    console.print(md)


def get_app():
    cli_app = typer.Typer(help="🔄 [s] self operations subcommands", no_args_is_help=True, add_help_option=True, add_completion=False)
    cli_app.command("update",      no_args_is_help=False, help="🔄 [u] UPDATE machineconfig")(update)
    cli_app.command("u",           no_args_is_help=False, hidden=True)(update)
    cli_app.command("interactive", no_args_is_help=False, help="🤖 [ia] INTERACTIVE configuration of machine.")(interactive)
    cli_app.command("ia",           no_args_is_help=False, help="INTERACTIVE configuration of machine.", hidden=True)(interactive)
    cli_app.command(name="init",         no_args_is_help=False, help="🦐 [t] Define and manage configurations")(init)
    cli_app.command(name="t",            no_args_is_help=False, hidden=True)(init)
    cli_app.command("status",      no_args_is_help=False, help="📊 [s] STATUS of machine, shell profile, apps, symlinks, dotfiles, etc.")(status)
    cli_app.command("s",           no_args_is_help=False, help="STATUS of machine, shell profile, apps, symlinks, dotfiles, etc.", hidden=True)(status)
    cli_app.command("install",     no_args_is_help=False, help="📋 [i] CLONE machienconfig locally and incorporate to shell profile for faster execution and nightly updates.")(install)
    cli_app.command("i",           no_args_is_help=False, help="CLONE machienconfig locally and incorporate to shell profile for faster execution and nightly updates.", hidden=True)(install)
    cli_app.command("navigate", no_args_is_help=False, help="📚 [n] NAVIGATE command structure with TUI")(navigate)
    cli_app.command("n", no_args_is_help=False, help="NAVIGATE command structure with TUI", hidden=True)(navigate)

    cli_app.command("readme", no_args_is_help=False, help="📚 [r] render readme markdown in terminal.")(readme)
    cli_app.command("r", no_args_is_help=False, hidden=True)(readme)
    return cli_app




"""ID

On windows:

# on the side of the local machine, you need to sort out the following:
# if the $sshfile doesn't have a standard name, you will need to explicitly specify the identity while SSHing (e.g. ssh u@s -i ~/.ssh/my_id)
# However, this must be done every time. For permanent solutions, use .ssh/config
$sshfile = "$env:USERPROFILE/.ssh/id_rsa"
Set-Service ssh-agent -StartupType Manual  # allow the service to be started manually
ssh-agent  # start the service
ssh-add.exe $sshfile # add the key to the agent

# add key:
$sshfile=""
$ErrorActionPreference = "Stop"
$sshd_dir = "$env:ProgramData\ssh"
cp "$sshd_dir\administrators_authorized_keys" "$sshd_dir\administrators_authorized_keys.orig"
Get-Content $sshfile >> "$sshd_dir\administrators_authorized_keys"
Restart-Service sshd -Force

# copy ssh key:
# This is the Windows equivalent of copy-ssh-id on Linux.
# Just like the original function, it is a convenient way of doing two things in one go:
# 1- copy a certain public key to the remote machine.
#    scp ~/.ssh/id_rsa.pub $remote_user@$remote_host:~/.ssh/authorized_keys
# 2- Store the value on the remote in a file called .ssh/authorized_keys
#    ssh $remote_user@$remote_host "echo $public_key >> ~/.ssh/authorized_keys"
# Idea from: https://www.chrisjhart.com/Windows-10-ssh-copy-id/

$key_value = cat ($env:USERPROFILE + "/.ssh/id_rsa.pub")
ssh $args[0] "powershell.exe -Command type $key_value >> .ssh/authorized_keys"
$my_keys='https://github.com/thisismygitrepo.keys'
(Invoke-WebRequest $my_keys).Content >> .ssh/authorized_keys

"""

from machineconfig.utils.path_extended import PathExtended
from machineconfig.utils.options import choose_from_options
from rich.console import Console
from rich.panel import Panel


console = Console()


def main() -> None:
    private_keys = [PathExtended(x).with_name(x.stem) for x in PathExtended.home().joinpath(".ssh").glob("*.pub")]
    key_status = f"Found {len(private_keys)} key(s)" if private_keys else "No keys found"
    console.print(Panel(f"🔑 SSH Identity Management\n🔍 {key_status} in ~/.ssh", title="[bold blue]Setup[/bold blue]", expand=False))

    choice = choose_from_options(msg="Path to private key to be used when ssh'ing: ", options=[str(x) for x in private_keys] + ["I have the path to the key file", "I want to paste the key itself"], multi=False)

    if choice == "I have the path to the key file":
        path_to_key = PathExtended(input("📋 Enter path to private key: ")).expanduser().absolute()
    elif choice == "I want to paste the key itself":
        key_filename = input("📝 File name (default: my_pasted_key): ") or "my_pasted_key"
        path_to_key = PathExtended.home().joinpath(f".ssh/{key_filename}")
        path_to_key.parent.mkdir(parents=True, exist_ok=True)
        path_to_key.write_text(input("🔑 Paste the private key: "), encoding="utf-8")
    else:
        path_to_key = PathExtended(choice)

    txt = f"IdentityFile {path_to_key.collapseuser().as_posix()}"
    config_path = PathExtended.home().joinpath(".ssh/config")

    if config_path.exists():
        current = config_path.read_text(encoding="utf-8")
        config_action = "updated"
    else:
        current = txt
        config_action = "created"
    lines = current.split("\n")
    found = False
    for i, line in enumerate(lines):
        if txt in line:
            lines[i] = txt
            found = True
    if not found:
        lines.insert(0, txt)
    config_path.write_text("\n".join(lines), encoding="utf-8")

    console.print(Panel(f"✅ Identity: {path_to_key.name}\n📄 Config {config_action}: {config_path}", title="[bold green]Complete[/bold green]", expand=False, border_style="green"))
    return None


if __name__ == "__main__":
    pass



"""Pure Python implementations for sessions commands - no typer dependencies."""

from typing import Optional, Literal
from pathlib import Path


def balance_load(
    layout_path: str,
    max_thresh: int,
    thresh_type: Literal["number", "n", "weight", "w"],
    breaking_method: Literal["moreLayouts", "ml", "combineTabs", "ct"],
    output_path: Optional[str],
) -> None:
    """Adjust layout file to limit max tabs per layout, etc."""
    thresh_type_resolved: dict[str, Literal["number", "weight"]] = {"number": "number", "n": "number", "weight": "weight", "w": "weight"}
    breaking_method_resolved: dict[str, Literal["moreLayouts", "combineTabs"]] = {"moreLayouts": "moreLayouts", "ml": "moreLayouts", "combineTabs": "combineTabs", "ct": "combineTabs"}

    layout_path_obj = Path(layout_path).expanduser().absolute()

    from machineconfig.utils.schemas.layouts.layout_types import LayoutsFile
    import json
    layoutfile: LayoutsFile = json.loads(layout_path_obj.read_text())
    layout_configs = layoutfile["layouts"]
    from machineconfig.cluster.sessions_managers.utils.load_balancer import limit_tab_num
    new_layouts = limit_tab_num(layout_configs=layout_configs, max_thresh=max_thresh, threshold_type=thresh_type_resolved[thresh_type], breaking_method=breaking_method_resolved[breaking_method])
    layoutfile["layouts"] = new_layouts
    target_file = Path(output_path) if output_path is not None else layout_path_obj.parent / f"{layout_path_obj.stem}_adjusted_{max_thresh}_{thresh_type}_{breaking_method}.json"
    target_file.parent.mkdir(parents=True, exist_ok=True)
    target_file.write_text(data=json.dumps(layoutfile, indent=4), encoding="utf-8")
    print(f"Adjusted layout saved to {target_file}")


def select_layout(layouts_json_file: str, selected_layouts_names: Optional[list[str]], select_interactively: bool, subsitute_home: bool) -> list["LayoutConfig"]:  # type: ignore[name-defined]
    """Select layout(s) from a layout file."""
    import json
    from machineconfig.utils.schemas.layouts.layout_types import LayoutsFile, LayoutConfig
    json_str = Path(layouts_json_file).read_text(encoding="utf-8")
    if subsitute_home:
        json_str = json_str.replace("~", str(Path.home())).replace("$HOME", str(Path.home()))
        json_str = json_str.replace("""Command": "f """, """Command": "~/.config/machineconfig/scripts/wrap_mcfg fire """)
        json_str = json_str.replace("""Command": "s """, """Command": "~/.config/machineconfig/scripts/wrap_mcfg sessions """)

    layout_file: LayoutsFile = json.loads(json_str)
    if len(layout_file["layouts"]) == 0:
        raise ValueError(f"No layouts found in {layouts_json_file}")
    if selected_layouts_names is None:
        if not select_interactively:
            return layout_file["layouts"]
        options = [layout["layoutName"] for layout in layout_file["layouts"]]
        from machineconfig.utils.options import choose_from_options
        selected_layouts_names = choose_from_options(multi=True, options=options, prompt="Choose a layout configuration:", tv=True, msg="Choose one option")
    print(f"Selected layout(s): {selected_layouts_names}")
    layouts_chosen: list[LayoutConfig] = []
    for name in selected_layouts_names:
        layout_chosen = next((layout for layout in layout_file["layouts"] if layout["layoutName"] == name), None)
        if layout_chosen is None:
            layout_chosen = next((layout for layout in layout_file["layouts"] if layout["layoutName"].lower() == name.lower()), None)
        if layout_chosen is None:
            available_layouts = [layout["layoutName"] for layout in layout_file["layouts"]]
            raise ValueError(f"Layout '{name}' not found. Available layouts: {available_layouts}")
        layouts_chosen.append(layout_chosen)
    return layouts_chosen


def find_layout_file(layout_path: str) -> str:
    """Find layout file from a path."""
    from machineconfig.utils.path_helper import search_for_files_of_interest, match_file_name, sanitize_path
    from machineconfig.utils.options import choose_from_options
    path_obj = sanitize_path(layout_path)
    if not path_obj.exists():
        choice_file = match_file_name(sub_string=layout_path, search_root=Path.cwd(), suffixes={".json"})
    elif path_obj.is_dir():
        print(f"🔍 Searching recursively for Python, PowerShell and Shell scripts in directory `{path_obj}`")
        files = search_for_files_of_interest(path_obj, suffixes={".py", ".sh", ".ps1"})
        print(f"🔍 Got #{len(files)} results.")
        choice_file = choose_from_options(multi=False, options=files, tv=True, msg="Choose one option")
        choice_file = Path(choice_file).expanduser().absolute()
    else:
        choice_file = path_obj
    return str(choice_file)


def run_layouts(
    layout_path: Optional[str],
    max_tabs: int,
    max_layouts: int,
    sleep_inbetween: float,
    monitor: bool,
    parallel: bool,
    kill_upon_completion: bool,
    choose: Optional[str],
    choose_interactively: bool,
    subsitute_home: bool,
) -> None:
    """Launch terminal sessions based on a layout configuration file."""
    if layout_path is None:
        raise ValueError("layout_path is required")

    layout_path_resolved = find_layout_file(layout_path=layout_path)
    layouts_selected = select_layout(layouts_json_file=layout_path_resolved, selected_layouts_names=choose.split(",") if choose else None, select_interactively=choose_interactively, subsitute_home=subsitute_home)

    if parallel and len(layouts_selected) > max_layouts:
        raise ValueError(f"Number of layouts {len(layouts_selected)} exceeds the maximum allowed {max_layouts}. Please adjust your layout file.")
    for a_layout in layouts_selected:
        if len(a_layout["layoutTabs"]) > max_tabs:
            raise ValueError(f"Layout '{a_layout.get('layoutName', 'Unnamed')}' has {len(a_layout['layoutTabs'])} tabs which exceeds the max of {max_tabs}.")

    import time
    import platform
    if platform.system() == "Linux" or platform.system() == "Darwin":
        from machineconfig.cluster.sessions_managers.zellij_local_manager import ZellijLocalManager
        if not parallel:
            iterable = [[item] for item in layouts_selected]
        else:
            iterable = [layouts_selected]
        for i, a_layouts in enumerate(iterable):
            manager = ZellijLocalManager(session_layouts=a_layouts)
            manager.start_all_sessions(poll_interval=2, poll_seconds=2)
            if monitor:
                manager.run_monitoring_routine(wait_ms=2000)
                if kill_upon_completion:
                    manager.kill_all_sessions()
            if i < len(layouts_selected) - 1:
                time.sleep(sleep_inbetween)
    elif platform.system() == "Windows":
        from machineconfig.cluster.sessions_managers.wt_local_manager import WTLocalManager
        if not parallel:
            iterable = [[item] for item in layouts_selected]
        else:
            iterable = [layouts_selected]
        for i, a_layouts in enumerate(iterable):
            manager = WTLocalManager(session_layouts=a_layouts)
            manager.start_all_sessions()
            if monitor:
                manager.run_monitoring_routine(wait_ms=2000)
                if kill_upon_completion:
                    manager.kill_all_sessions()
            if i < len(layouts_selected) - 1:
                time.sleep(sleep_inbetween)
    else:
        print(f"❌ Unsupported platform: {platform.system()}")


def create_template(name: Optional[str], num_tabs: int) -> None:
    """Create a layout template file."""
    from machineconfig.utils.schemas.layouts.layout_types import LayoutsFile, TabConfig, LayoutConfig
    tabs: list[TabConfig] = []
    for i in range(1, num_tabs + 1):
        tab: TabConfig = {
            "tabName": f"Tab{i}",
            "startDir": "~/" + str(Path.cwd().relative_to(Path.home())),
            "command": "bash",
        }
        tabs.append(tab)
    layouts: list[LayoutConfig] = [
        {
            "layoutName": f"{Path.cwd().name}Layout",
            "layoutTabs": tabs,
        }
    ]
    file: LayoutsFile = {
        "$schema": "https://bit.ly/cfglayout",  # type: ignore
        "version": "0.1",
        "layouts": layouts,
    }
    import json
    json_string = json.dumps(file, indent=4)
    if name is None:
        layout_path = Path.cwd() / "layout.json"
    else:
        layout_path = Path.cwd() / (name.replace(".json", "") + ".json")
    layout_path.parent.mkdir(parents=True, exist_ok=True)
    if layout_path.exists():
        print(f"❌ File {layout_path} already exists. Aborting to avoid overwriting.")
        return
    layout_path.write_text(json_string, encoding="utf-8")
    print(f"✅ Created layout template at {layout_path}")





import platform
from typing import Optional
from rich import box
from rich.console import Console
from rich.panel import Panel
from machineconfig.utils.schemas.installer.installer_types import InstallerData

ps1 = r"""
$winget = Get-Command winget -ErrorAction SilentlyContinue
if (-not $winget) {
    Write-Host "winget not found. Installing..."
    $finalUrl = (Invoke-WebRequest 'https://github.com/microsoft/winget-cli/releases/latest' -UseBasicParsing).BaseResponse.ResponseUri.AbsoluteUri
    $releaseTag = $finalUrl.Split('/')[-1]
    $DownloadUrl = "https://github.com/microsoft/winget-cli/releases/download/$releaseTag/Microsoft.DesktopAppInstaller_8wekyb3d8bbwe.msixbundle"
    $DestDir     = Join-Path $HOME "Downloads"
    $DestFile    = Join-Path $DestDir "Microsoft.DesktopAppInstaller_8wekyb3d8bbwe.msixbundle"
    # Create folder if it doesn't exist
    if (-not (Test-Path $DestDir)) {
        New-Item -ItemType Directory -Path $DestDir | Out-Null
    }
    Write-Host "Downloading winget installer..."
    # Invoke-WebRequest -Uri $DownloadUrl -OutFile $DestFile
    # Start-BitsTransfer -Source $DownloadUrl -Destination $DestFile  # 
    #Invoke-WebRequest -Uri $DownloadUrl -OutFile $DestFile -UseBasicParsing
    curl.exe -L -o $DestFile $DownloadUrl
    Write-Host "Saved to: $DestFile"
    # We MUST run Add-AppxPackage in Windows PowerShell
    Write-Host "Installing package via Windows PowerShell..."
    powershell.exe -NoLogo -NoProfile -Command "Add-AppxPackage -Path `"$DestFile`" "
    Write-Host "Installation complete."
    }
else {
    Write-Host "winget already available. Skipping installation."
}

winget install --no-upgrade --name "Powershell"                   --Id "Microsoft.PowerShell"       --source winget --scope user --accept-package-agreements --accept-source-agreements  # powershell require admin
winget install --no-upgrade --name "Windows Terminal"             --Id "Microsoft.WindowsTerminal"  --source winget --scope user --accept-package-agreements --accept-source-agreements  # Terminal is is installed by default on W 11
# --GROUP:gui:Brave+VSCode+Git+WezTerm
# --GROUP:dev2:VSRedistrib+VSBuildTools+Codeblocks+GnuWin32: Make+GnuPG+graphviz+WinFsp+SSHFS-win+xming+Node.js+Rustup+Cloudflare+Cloudflare WARP+Microsoft Garage Mouse without Borders
# --GROUP:user:nu+Chrome+ChromeRemoteDesktop+Zoom+7zip+Firefox+Thunderbird+StreamlabsOBS+OBSStudio+MiKTeX+TexMaker+notepad+++Lapce+TesseractOCR+perl+DB Browser for SQLite+sql server management studio+Adobe Acrobat Reader DC+julia+Chafa+bottom+onefetch+Just+hyperfine+AWS CLI
# Install-Module -Name Terminal-Icons -Repository PSGallery -Force -AcceptLicense -PassThru -Confirm  # -RequiredVersion 2.5.10
# [System.Environment]::SetEnvironmentVariable('PYTHONUTF8', '1', 'User')
# [System.Environment]::SetEnvironmentVariable('PYTHONIOENCODING', 'utf-8', 'User')


"""

zsh = r"""
/bin/bash -c "$(curl -fsSL https://raw.githubusercontent.com/Homebrew/install/HEAD/install.sh)"
echo "🔄 Updating Homebrew..."
brew update || true
# Note: git and nano are pre-installed on macOS, but we install via Homebrew to ensure latest versions
# brew install git || true
# brew install nano || true
# brew install curl || true
# Install NVM
if [ ! -s "$HOME/.nvm/nvm.sh" ]; then
    echo "📥 Installing NVM (Node Version Manager)..."
    curl -o- https://raw.githubusercontent.com/nvm-sh/nvm/v0.40.3/install.sh | bash
fi
echo "🔧 Configuring NVM environment..."
export NVM_DIR="$HOME/.nvm"
[ -s "$NVM_DIR/nvm.sh" ] && \. "$NVM_DIR/nvm.sh"
echo "📥 Installing latest Node.js..."
nvm install node || true
# brew install make
# brew install ffmpeg
# brew install openssl
echo "✅ Essential tools installation complete."
"""

bash = r"""
sudo apt update -y || true
sudo apt install nala -y || true
sudo nala install curl wget gpg lsb-release apt-transport-https -y || true
sudo nala install git net-tools htop nano -y || true
sudo nala install build-essential python3-dev -y || true  # C build toolchain: Where build-essential brings gcc, make, etc., and python3-dev ensures headers for your Python version.
# sudo nala install libssl-dev -y
# sudo nala install libaa-bin -y

curl -o- https://raw.githubusercontent.com/nvm-sh/nvm/v0.40.3/install.sh | bash
source ~/.bashrc || true
nvm install node || true

sudo nala install samba -y || true
sudo nala install fuse3 -y || true
sudo nala install nfs-common -y || true

# echo 'keyboard-configuration keyboard-configuration/layout select US English' | sudo debconf-set-selections
# echo 'keyboard-configuration keyboard-configuration/layoutcode string us' | sudo debconf-set-selections
# sudo DEBIAN_FRONTEND=noninteractive nala install -y cmatrix
# sudo nala install hollywood -y || true

# sudo nala install ffmpeg -y || true  # Required by some dev tools
# sudo nala install make -y || true  # Required by LunarVim and SpaceVim
# (curl --proto '=https' --tlsv1.2 -sSf https://sh.rustup.rs | sh) || true

"""


def main(installer_data: InstallerData, version: Optional[str]) -> None:
    console = Console()
    _ = installer_data
    console.print(
        Panel.fit(
            "\n".join([f"💻 Platform: {platform.system()}", f"🔄 Version: {'latest' if version is None else version}"]),
            title="🔧 ABC Installer",
            border_style="blue",
            box=box.ROUNDED,
        )
    )

    _ = version
    if platform.system() == "Windows":
        console.print("🪟 Installing ABC on Windows using winget...", style="bold")
        program = ps1
    elif platform.system() == "Linux":
        console.print("🐧 Installing ABC on Linux...", style="bold")
        program = bash
    elif platform.system() == "Darwin":
        console.print("🍎 Installing ABC on macOS...", style="bold")
        program = zsh
    else:
        error_msg = f"Unsupported platform: {platform.system()}"
        console.print(
            Panel.fit(
                "\n".join([error_msg]),
                title="❌ Error",
                subtitle="⚠️ Unsupported platform",
                border_style="red",
                box=box.ROUNDED,
            )
        )
        raise NotImplementedError(error_msg)
    from machineconfig.utils.code import print_code, run_shell_script
    print_code(code=program, lexer="shell", desc="Installation Script Preview")
    run_shell_script(program)





from typing import Optional
from pathlib import Path
from machineconfig.utils.accessories import randstr
from machineconfig.utils.meta import lambda_to_python_script
from machineconfig.utils.ssh_utils.abc import DEFAULT_PICKLE_SUBDIR
from machineconfig.utils.code import get_uv_command


def copy_from_here(
    self: "SSH", source_path: str, target_rel2home: Optional[str], compress_with_zip: bool, recursive: bool, overwrite_existing: bool
) -> None:
    if self.sftp is None:
        raise RuntimeError(f"SFTP connection not available for {self.hostname}. Cannot transfer files.")
    source_obj = Path(source_path).expanduser().absolute()
    if not source_obj.exists():
        raise RuntimeError(f"SSH Error: source `{source_obj}` does not exist!")
    if target_rel2home is None:
        try:
            target_rel2home = str(source_obj.relative_to(Path.home()))
        except ValueError:
            raise RuntimeError(f"If target is not specified, source must be relative to home directory, but got: {source_obj}")
    if not compress_with_zip and source_obj.is_dir():
        if not recursive:
            raise RuntimeError(
                f"SSH Error: source `{source_obj}` is a directory! Set `recursive=True` for recursive sending or `compress_with_zip=True` to zip it first."
            )
        file_paths_to_upload: list[Path] = [file_path for file_path in source_obj.rglob("*") if file_path.is_file()]
        self.create_parent_dir_and_check_if_exists(path_rel2home=target_rel2home, overwrite_existing=overwrite_existing)
        for idx, file_path in enumerate(file_paths_to_upload):
            print(f"   {idx + 1:03d}. {file_path}")
        for file_path in file_paths_to_upload:
            remote_file_target = Path(target_rel2home).joinpath(file_path.relative_to(source_obj))
            self.copy_from_here(
                source_path=str(file_path),
                target_rel2home=str(remote_file_target),
                compress_with_zip=False,
                recursive=False,
                overwrite_existing=overwrite_existing,
            )
        return None
    if compress_with_zip:
        print("🗜️ ZIPPING ...")
        import shutil

        zip_path = Path(str(source_obj) + "_archive")
        if source_obj.is_dir():
            shutil.make_archive(str(zip_path), "zip", source_obj)
        else:
            shutil.make_archive(str(zip_path), "zip", source_obj.parent, source_obj.name)
        source_obj = Path(str(zip_path) + ".zip")
        if not target_rel2home.endswith(".zip"):
            target_rel2home = target_rel2home + ".zip"
    if Path(target_rel2home).parent.as_posix() not in {"", "."}:
        self.create_parent_dir_and_check_if_exists(path_rel2home=target_rel2home, overwrite_existing=overwrite_existing)
    print(f"""📤 [SFTP UPLOAD] Sending file: {repr(source_obj)}  ==>  Remote Path: {target_rel2home}""")
    try:
        with self.tqdm_wrap(ascii=True, unit="b", unit_scale=True) as pbar:
            if self.sftp is None:  # type: ignore[unreachable]
                raise RuntimeError(f"SFTP connection lost for {self.hostname}")
            print(f"Uploading {source_obj} to\n{Path(self.remote_specs['home_dir']).joinpath(target_rel2home)}")
            self.sftp.put(
                localpath=str(source_obj), remotepath=str(Path(self.remote_specs["home_dir"]).joinpath(target_rel2home)), callback=pbar.view_bar
            )
    except Exception:
        if compress_with_zip and source_obj.exists() and str(source_obj).endswith("_archive.zip"):
            source_obj.unlink()
        raise

    if compress_with_zip:

        def unzip_archive(zip_file_path: str, overwrite_flag: bool) -> None:
            from pathlib import Path
            import shutil
            import zipfile

            archive_path = Path(zip_file_path).expanduser()
            extraction_directory = archive_path.parent / archive_path.stem
            if overwrite_flag and extraction_directory.exists():
                shutil.rmtree(extraction_directory)
            with zipfile.ZipFile(archive_path, "r") as archive_handle:
                archive_handle.extractall(extraction_directory)
            archive_path.unlink()

        command = lambda_to_python_script(
            lambda: unzip_archive(
                zip_file_path=str(Path(self.remote_specs["home_dir"]).joinpath(target_rel2home)), overwrite_flag=overwrite_existing
            ),
            in_global=True,
            import_module=False,
        )
        tmp_py_file = Path.home().joinpath(f"{DEFAULT_PICKLE_SUBDIR}/create_target_dir_{randstr()}.py")
        tmp_py_file.parent.mkdir(parents=True, exist_ok=True)
        tmp_py_file.write_text(command, encoding="utf-8")
        remote_tmp_py = tmp_py_file.relative_to(Path.home()).as_posix()
        self.copy_from_here(source_path=str(tmp_py_file), target_rel2home=None, compress_with_zip=False, recursive=False, overwrite_existing=True)
        self.run_shell_cmd_on_remote(
            command=f"""{get_uv_command(platform=self.remote_specs['system'])} run python {remote_tmp_py}""",
            verbose_output=False,
            description=f"UNZIPPING {target_rel2home}",
            strict_stderr=True,
            strict_return_code=True,
        )
        source_obj.unlink()
        tmp_py_file.unlink(missing_ok=True)
    return None


if __name__ == "__main__":
    from machineconfig.utils.ssh import SSH





from typing import Annotated, Optional
import typer
from pathlib import Path


def download(
    url: Annotated[Optional[str], typer.Argument(..., help="The URL to download the file from.")] = None,
    decompress: Annotated[bool, typer.Option(..., "--decompress", "-d", help="Decompress the file if it's an archive.")] = False,
    output: Annotated[Optional[str], typer.Option("--output", "-o", help="The output file path.")] = None,
    output_dir: Annotated[Optional[str], typer.Option("--output-dir", help="Directory to place the downloaded file in.")] = None,
) -> Optional["Path"]:
    import subprocess
    from urllib.parse import parse_qs, unquote, urlparse
    from requests import Response
    import requests
    from pathlib import Path
    if url is None:
        typer.echo("❌ Error: URL is required.", err=True)
        return None
    if output is not None and output_dir is not None:
        typer.echo("❌ Error: --output and --output-dir cannot be used together.", err=True)
        return None
    typer.echo(f"📥 Downloading from: {url}")

    def _sanitize_candidate_filename(name: str) -> Optional[str]:
        candidate = Path(name).name.strip()
        if not candidate or candidate in {".", ".."}:
            return None
        return candidate

    def _filename_from_content_disposition(header_value: Optional[str]) -> Optional[str]:
        if header_value is None:
            return None
        parts = [segment.strip() for segment in header_value.split(";")]
        for part in parts:
            lower = part.lower()
            if lower.startswith("filename*="):
                value = part.split("=", 1)[1]
                value = value.strip().strip('"')
                if "''" in value:
                    value = value.split("''", 1)[1]
                decoded = unquote(value)
                sanitized = _sanitize_candidate_filename(decoded)
                if sanitized is not None:
                    return sanitized
            if lower.startswith("filename="):
                value = part.split("=", 1)[1].strip().strip('"')
                decoded = unquote(value)
                sanitized = _sanitize_candidate_filename(decoded)
                if sanitized is not None:
                    return sanitized
        return None

    def _filename_from_url(source_url: str) -> Optional[str]:
        parsed = urlparse(source_url)
        url_candidate = _sanitize_candidate_filename(unquote(Path(parsed.path).name))
        if url_candidate is not None:
            return url_candidate
        query_params = parse_qs(parsed.query, keep_blank_values=True)
        for key, values in query_params.items():
            lower_key = key.lower()
            if "name" in lower_key or "file" in lower_key:
                for value in values:
                    sanitized = _sanitize_candidate_filename(unquote(value))
                    if sanitized is not None:
                        return sanitized
        return None

    def _resolve_download_path(request_url: str, response: Response, requested_output: Optional[str], requested_output_dir: Optional[str]) -> Path:
        if requested_output is not None:
            return Path(requested_output)
        header_candidate = _filename_from_content_disposition(response.headers.get("content-disposition"))
        if header_candidate is None:
            header_candidate = _filename_from_url(response.url)
        if header_candidate is None:
            header_candidate = _filename_from_url(request_url)
        if header_candidate is None:
            header_candidate = "downloaded_file"
        if requested_output_dir is not None:
            return Path(requested_output_dir) / header_candidate
        return Path(header_candidate)

    try:
        with requests.get(url, allow_redirects=True, stream=True, timeout=60) as response:
            response.raise_for_status()
            download_path = _resolve_download_path(url, response, output, output_dir)
            download_path.parent.mkdir(parents=True, exist_ok=True)
            total_size_header = response.headers.get("content-length", "0")
            try:
                total_size = int(total_size_header)
            except (TypeError, ValueError):
                total_size = 0
            if total_size <= 0:
                with open(download_path, "wb") as file_handle:
                    file_handle.write(response.content)
            else:
                downloaded = 0
                chunk_size = 8192 * 40
                with open(download_path, "wb") as file_handle:
                    for chunk in response.iter_content(chunk_size=chunk_size):
                        if not chunk:
                            continue
                        file_handle.write(chunk)
                        downloaded += len(chunk)
                        progress = (downloaded / total_size) * 100
                        typer.echo(f"\r⏬ Progress: {progress:.1f}% ({downloaded}/{total_size} bytes)", nl=False)
                typer.echo()
    except requests.exceptions.RequestException as exception:
        typer.echo(f"❌ Download failed: {exception}", err=True)
        return None
    except OSError as exception:
        typer.echo(f"❌ File write error: {exception}", err=True)
        return None

    typer.echo(f"✅ Downloaded to: {download_path}")
    result_path: Path = download_path
    if decompress:
        typer.echo(f"📦 Decompressing: {download_path}")
        base_name = download_path.name.split(".", maxsplit=1)[0]  # ouch decompresses all (e.g. .tar.gz) in one go.
        if base_name in {"", ".", ".."}:
            base_name = "extracted"
        extract_dir = download_path.parent / base_name
        extract_dir.mkdir(parents=True, exist_ok=True)
        try:
            subprocess.run(
                ["ouch", "decompress", str(download_path), "--dir", str(extract_dir)],
                check=True,
                capture_output=True,
                text=True,
            )
            typer.echo(f"✅ Decompressed to: {extract_dir}")
            if download_path.exists():
                download_path.unlink()
                typer.echo(f"🗑️  Removed archive: {download_path}")
            result_path = extract_dir
        except subprocess.CalledProcessError as exception:
            typer.echo(f"❌ Decompression failed: {exception.stderr}", err=True)
            return None
        except FileNotFoundError:
            typer.echo("❌ Error: ouch command not found. Please install ouch.", err=True)
            typer.echo("💡 Install with: cargo install ouch", err=True)
            return None

    return result_path.resolve()


if __name__ == "__main__":
    pass




from typing import Optional, TypedDict, cast

class CountryFlag(TypedDict, total=False):
    emoji: str
    unicode: str

class CountryCurrency(TypedDict, total=False):
    code: str
    symbol: str

class Continent(TypedDict, total=False):
    code: str
    name: str

class PublicIpInfo(TypedDict, total=True):
    ip: str
    hostname: str
    city: str
    region: str
    country: str
    country_name: str
    country_flag: CountryFlag
    country_flag_url: str
    country_currency: CountryCurrency
    continent: Continent
    loc: str
    org: str
    postal: str
    timezone: str


def get_public_ip_address() -> PublicIpInfo:
    from machineconfig.utils.installer_utils.installer_cli import install_if_missing
    import subprocess
    install_if_missing("ipinfo")
    result = subprocess.run(
        ["ipinfo", "myip", "--json"],
        check=True,
        capture_output=True,
        text=True,
        encoding="utf-8",
    )
    import json
    loaded_json: PublicIpInfo = json.loads(result.stdout)
    return loaded_json


def get_all_ipv4_addresses() -> list[tuple[str, str]]:
    import psutil
    import socket
    result: list[tuple[str, str]] = []
    for iface, addrs in psutil.net_if_addrs().items():
        for addr in addrs:
            if addr.family == socket.AF_INET:
                ip = addr.address
                result.append((iface, ip))
    return result


def select_lan_ipv4(prefer_vpn: bool) -> Optional[str]:
    """
    Choose the best 'real LAN' IPv4:
      - Excludes loopback/link-local and (by default) VPN/tunnel/container ifaces
      - Prefers physical-looking ifaces (eth/en*/wlan/wl*)
      - Prefers RFC1918 LANs: 192.168/16 > 10/8 > 172.16/12
      - Requires interface is UP
    Set prefer_vpn=True to allow tunnel/VPN ifaces to compete.
    """

    import ipaddress
    import re
    from collections.abc import Sequence
    import psutil

    # Down-rank or exclude: tunnels/VPNs/bridges/containers (add your own if needed)
    VIRTUAL_IFACE_PAT = re.compile(
        r"^(?:lo|loopback|docker\d*|br-.*|veth.*|virbr.*|bridge.*|"
        r"vboxnet.*|vmnet.*|zt.*|ham.*|tailscale.*|wg\d*|utun\d*|llw\d*|awdl\d*|"
        r"tun\d*|tap\d*|cloudflarewarp.*|warp.*)$",
        re.IGNORECASE,
    )

    # Light preference for names that look like real NICs
    PHYSICAL_IFACE_PAT = re.compile(
        r"^(?:eth\d*|en\d*|enp.*|ens.*|eno.*|wlan\d*|wl.*|.*wifi.*|.*ethernet.*)$",
        re.IGNORECASE,
    )

    # Known noisy CIDRs to avoid
    NOISY_NETS: list[ipaddress.IPv4Network] = [
        ipaddress.IPv4Network("100.64.0.0/10"),   # CGNAT (Tailscale/others)
        ipaddress.IPv4Network("172.17.0.0/16"),   # docker0 default
        ipaddress.IPv4Network("172.18.0.0/16"),
        ipaddress.IPv4Network("172.19.0.0/16"),
        ipaddress.IPv4Network("192.168.49.0/24"), # minikube default
        ipaddress.IPv4Network("10.0.2.0/24"),     # VirtualBox NAT
    ]

    def _in_any(ip: ipaddress.IPv4Address, nets: Sequence[ipaddress.IPv4Network]) -> bool:
        return any(ip in n for n in nets)

    stats = psutil.net_if_stats()
    best = None
    best_score = -10**9
    import socket
    for iface, addrs in psutil.net_if_addrs().items():
        st = stats.get(iface)
        if not st or not st.isup:
            continue

        for a in addrs:
            if a.family != socket.AF_INET or not a.address:
                continue

            ip_str = a.address
            try:
                ip = cast(ipaddress.IPv4Address, ipaddress.ip_address(ip_str))
            except ValueError:
                continue

            # Exclude unusable classes
            if ip.is_loopback or ip.is_link_local:  # 127.0.0.0/8, 169.254.0.0/16
                continue

            # Hard filter: if it looks virtual and we don't prefer VPNs, skip it
            if not prefer_vpn and VIRTUAL_IFACE_PAT.match(iface):
                continue

            # Hard filter: known noisy subnets (docker, cgnat, etc.)
            if _in_any(ip, NOISY_NETS) and not prefer_vpn:
                continue

            # Base score
            score = 0

            # Prefer physical-looking names
            if PHYSICAL_IFACE_PAT.match(iface):
                score += 200

            # Broadcast present usually means L2 LAN (not point-to-point)
            # (psutil puts it on the same entry as .broadcast)
            if getattr(a, "broadcast", None):
                score += 100

            # Prefer private RFC1918; rank families
            if ip.is_private:
                # Order: 192.168.x.x > 10.x.x.x > 172.16-31.x.x
                ip_net = ipaddress.IPv4Network((ip, 32), strict=False)
                if ipaddress.IPv4Network("192.168.0.0/16").supernet_of(ip_net):
                    score += 90
                elif ipaddress.IPv4Network("10.0.0.0/8").supernet_of(ip_net):
                    score += 70
                elif ipaddress.IPv4Network("172.16.0.0/12").supernet_of(ip_net):
                    score += 50
            else:
                # Public on a NIC is unusual for a home/office LAN
                score -= 50

            # Slight nudge by interface speed if known (>0 means psutil knows it)
            # (Many tunnels report 0)
            if getattr(st, "speed", 0) > 0:
                score += 20

            # Deterministic tie-breaker: prefer shorter iface name (eth0 over eth10)
            score -= len(iface) * 0.01

            if score > best_score:
                best_score = score
                best = ip_str

    return best


if __name__ == "__main__":
    print(select_lan_ipv4(False) or "No LAN IPv4 found")



"""

create 18 tasks one per cli
in each one, you do comprehensive search on internet on how to set up the cli/extension or whatever so we get max privacy and security (nothing shared everything is declined and is on the safe side)
this is a moving space, everyday, there are new versions, new configs, new settings, new policies, new rules, new terms of service, new eula, new privacy policies etc etc
so please search for latest and don't be complacent with the work done so far.

do it one by one, don't research and implement next before finishing the task before it.

I provided gemini cli solution as a reference (feel free to make it more flexible if you need mroe room to implement something more complex than what I did, its just a guide)
remember that other cli's have different names and configs and locations etc, read official docs

in the following cli how to set it for maximal prviacy *turn off any telemetry, reject any use of data, etc etc, please think of everything and read all theird docs and fine print
    "aider",
    "aichat",  # rust cli
    "copilot",  # cli from github copilot extension
    "gemini",  # gemini cli 
    "crush",  # cli from charm team
    "mods",  # cli from same people as crush (not agent) @ https://github.com/charmbracelet/mods
    "opencode-ai",
    "chatgpt",
    "q",  # from aws
    "qwen-code",  # fork of gemini cli
    "cursor-cli",
    "droid",
    "kilocode",
    "cline",  # cli from cline ai extension
    "auggie",
    "codex",  # from openai

"""

from typing import Optional


def apply_max_privacy_and_security_rules_and_configs(overwrite: bool, repo_root: Optional[str]) -> None:
    from machineconfig.utils.source_of_truth import LIBRARY_ROOT
    root = LIBRARY_ROOT / "scripts/python/helpers_agents/privacy/configs"
    from pathlib import Path

    # Gemini privacy settings
    gemini_settings_source = root.joinpath("gemini/settings.json")
    gemini_settings_target_global = Path.home().joinpath(".gemini/settings.json")
    if not gemini_settings_target_global.exists() or overwrite:
        gemini_settings_target_global.parent.mkdir(parents=True, exist_ok=True)
        gemini_settings_target_global.write_text(gemini_settings_source.read_text())
    if repo_root:
        gemini_settings_target_repo = Path(repo_root).joinpath(".gemini/settings.json")
        if not gemini_settings_target_repo.exists() or overwrite:
            gemini_settings_target_repo.parent.mkdir(parents=True, exist_ok=True)
            gemini_settings_target_repo.write_text(gemini_settings_source.read_text())

    # Aider privacy settings
    aider_settings_source = root.joinpath("aider/.aider.conf.yml")
    aider_settings_target_global = Path.home().joinpath(".aider.conf.yml")
    if not aider_settings_target_global.exists() or overwrite:
        aider_settings_target_global.parent.mkdir(parents=True, exist_ok=True)
        aider_settings_target_global.write_text(aider_settings_source.read_text())

    if repo_root:
        aider_settings_target_repo = Path(repo_root).joinpath(".aider.conf.yml")
        if not aider_settings_target_repo.exists() or overwrite:
            aider_settings_target_repo.parent.mkdir(parents=True, exist_ok=True)
            aider_settings_target_repo.write_text(aider_settings_source.read_text())

    # Aichat privacy settings
    aichat_settings_source = root.joinpath("aichat/config.yaml")
    aichat_settings_target_global = Path.home().joinpath(".config/aichat/config.yaml")
    if not aichat_settings_target_global.exists() or overwrite:
        aichat_settings_target_global.parent.mkdir(parents=True, exist_ok=True)
        aichat_settings_target_global.write_text(aichat_settings_source.read_text())

    if repo_root:
        aichat_settings_target_repo = Path(repo_root).joinpath(".config/aichat/config.yaml")
        if not aichat_settings_target_repo.exists() or overwrite:
            aichat_settings_target_repo.parent.mkdir(parents=True, exist_ok=True)
            aichat_settings_target_repo.write_text(aichat_settings_source.read_text())

    # Copilot privacy settings
    copilot_settings_source = root.joinpath("copilot/config.yml")
    copilot_settings_target_global = Path.home().joinpath(".config/gh-copilot/config.yml")
    if not copilot_settings_target_global.exists() or overwrite:
        copilot_settings_target_global.parent.mkdir(parents=True, exist_ok=True)
        copilot_settings_target_global.write_text(copilot_settings_source.read_text())

    if repo_root:
        copilot_settings_target_repo = Path(repo_root).joinpath(".config/gh-copilot/config.yml")
        if not copilot_settings_target_repo.exists() or overwrite:
            copilot_settings_target_repo.parent.mkdir(parents=True, exist_ok=True)
            copilot_settings_target_repo.write_text(copilot_settings_source.read_text())

    # Crush privacy settings
    crush_settings_source = root.joinpath("crush/crush.json")
    crush_settings_target_global = Path.home().joinpath(".config/crush/crush.json")
    if not crush_settings_target_global.exists() or overwrite:
        crush_settings_target_global.parent.mkdir(parents=True, exist_ok=True)
        crush_settings_target_global.write_text(crush_settings_source.read_text())

    if repo_root:
        crush_settings_target_repo = Path(repo_root).joinpath(".crush.json")
        if not crush_settings_target_repo.exists() or overwrite:
            crush_settings_target_repo.parent.mkdir(parents=True, exist_ok=True)
            crush_settings_target_repo.write_text(crush_settings_source.read_text())

    # next, cursh cli


if __name__ == "__main__":
    pass



#!/usr/bin/env python3
"""
Interactive Machine Configuration Setup Script

A Python version of the interactive installation script that uses questionary
for better user experience with checkbox selections.


# echo # 📧 Thunderbird Setup Note:
# Run after installing Thunderbird and starting it once:
# cd ~/AppData/Roaming/ThunderBird/Profiles
# $res = ls
# $name = $res[0].Name
# mv $backup_folder $name


"""

import sys
from pathlib import Path
import platform
import questionary
from questionary import Choice
from rich.console import Console
from rich.panel import Panel
from rich.text import Text
from machineconfig.utils.code import run_shell_script

console = Console()


def display_header() -> None:
    from machineconfig.utils.installer_utils.installer_runner import get_machineconfig_version
    from rich.align import Align

    # Fancy ASCII art header
    ascii_art = """
    ╔═════════════════════════════════════════════════════════════════════════════════════════════════════════════╗
    ║                                                                                                             ║
    ║   ███╗   ███╗ █████╗  ██████╗██╗  ██╗██╗███╗   ██╗███████╗ ██████╗ ██████╗ ███╗   ██╗███████╗██╗ ██████╗    ║
    ║   ████╗ ████║██╔══██╗██╔════╝██║  ██║██║████╗  ██║██╔════╝██╔════╝██╔═══██╗████╗  ██║██╔════╝██║██╔════╝    ║
    ║   ██╔████╔██║███████║██║     ███████║██║██╔██╗ ██║█████╗  ██║     ██║   ██║██╔██╗ ██║█████╗  ██║██║  ███╗   ║
    ║   ██║╚██╔╝██║██╔══██║██║     ██╔══██║██║██║╚██╗██║██╔══╝  ██║     ██║   ██║██║╚██╗██║██╔══╝  ██║██║   ██║   ║
    ║   ██║ ╚═╝ ██║██║  ██║╚██████╗██║  ██║██║██║ ╚████║███████╗╚██████╗╚██████╔╝██║ ╚████║██║     ██║╚██████╔╝   ║
    ║   ╚═╝     ╚═╝╚═╝  ╚═╝ ╚═════╝╚═╝  ╚═╝╚═╝╚═╝  ╚═══╝╚══════╝ ╚═════╝ ╚═════╝ ╚═╝  ╚═══╝╚═╝     ╚═╝ ╚═════╝    ║
    ║                                                                                                             ║
    ╚═════════════════════════════════════════════════════════════════════════════════════════════════════════════╝
    """

    version = get_machineconfig_version()
    title = f"✨ MACHINE CONFIGURATION v{version} ✨"
    subtitle = "🎯 Your digital life manager. Dotfiles, data, code and more."
    bug_report = "🐛 Please report bugs to Alex Al-Saffar @ https://github.com/thisismygitrepo/machineconfig"

    # Print ASCII art
    console.print(Text(ascii_art, style="bold cyan"))
    console.print()
    
    # Print centered text elements
    console.print(Align.center(Text(title, style="bold bright_magenta")))
    console.print(Align.center(Text(subtitle, style="italic bright_blue")))
    console.print()
    console.print(Align.center(Text(bug_report, style="dim white")))
    console.print()
def display_completion_message() -> None:
    completion_text = Text("INSTALLATION COMPLETE", style="bold green")
    subtitle_text = Text("System setup finished successfully", style="italic green")
    console.print(Panel(f"✨ {completion_text}\n{subtitle_text}\n\n🎉 Your system has been configured successfully!\n🔄 You may need to reboot to apply all changes.", border_style="green", padding=(1, 2)))
def display_dotfiles_instructions() -> None:
    header_text = Text("DOTFILES MIGRATION", style="bold yellow")
    subtitle_text = Text("Configuration transfer options", style="italic yellow")
    instructions = """
    On remote, run:
    rm ~/dotfiles.zip || true
    ouch c ~/dotfiles dotfiles.zip --password rew
    # INSECURE OVER INTERNET: uvx wormhole-magic send ~/dotfiles.zip
    # LOCAL NETWORK:
    devops network share-server --no-auth ./dotfiles.zip
    On new machine, run:
    # INSECURE cd $HOME; uvx wormhole-magic receive dotfiles.zip --accept-file
    cd $HOME;
    rm ~/dotfiles.zip || true
    utils download $URL
    ouch d ~/dotfiles.zip

🖱️  [bold blue]Method 1: USING MOUSE WITHOUT KB OR BROWSER SHARE[/bold blue]
    On original machine, run:
    [dim]cd ~/dotfiles/creds/msc
    easy-sharing . --password rew --username al[/dim]
    Then open brave on new machine to get MouseWithoutBorders password

🔐 [bold blue]Method 2: USING SSH[/bold blue]
    FROM REMOTE, RUN:
    [dim]fptx ~/dotfiles $USER@$(hostname):^ -z
    # OR, using IP address if router has not yet found the hostname:
    fptx ~/dotfiles $USER@$(hostname -I | awk '{print $1}'):^ -z[/dim]

☁️  [bold blue]Method 3: USING INTERNET SECURE SHARE[/bold blue]
    [dim]cd ~
    cloud copy SHARE_URL . --config ss[/dim]
    (requires symlinks to be created first)
    """
    console.print(Panel(f"📂 {header_text}\n{subtitle_text}\n\n{instructions}", border_style="yellow", padding=(1, 2)))


def get_installation_choices() -> list[str]:
    """Get user choices for installation options."""
    choices = [
        Choice(value="install_machineconfig", title="🐍 Install machineconfig cli.", checked=False),
        Choice(value="sysabc", title="📥 Install System Package Manager (Needed for other apps to be installed).", checked=False),
        Choice(value="termabc", title="⚡ Install Terminal CLI apps essentials (group `termabc`)", checked=False),
        Choice(value="install_shell_profile", title="🐚 Configure Shell Profile And Map Other Configs.", checked=False),
        Choice(value="install_ssh_server", title="🔒 [ADVANCED] Configure SSH Server", checked=False),
        Choice(value="retrieve_repositories", title="📚 [ADVANCED] Retrieve Repositories", checked=False),
        Choice(value="retrieve_data", title="💾 [ADVANCED] Retrieve Data.", checked=False),
    ]
    selected = questionary.checkbox("Select the installation options you want to execute:", choices=choices, show_description=True).ask()
    return selected or []


def execute_installations(selected_options: list[str]) -> None:
    for maybe_a_group in selected_options:
        if maybe_a_group in ("termabc", "sysabc"):
            console.print(Panel("⚡ [bold bright_yellow]CLI APPLICATIONS[/bold bright_yellow]\n[italic]Command-line tools installation[/italic]", border_style="bright_yellow"))
            console.print("🔧 Installing CLI applications", style="bold cyan")
            try:
                from machineconfig.utils.installer_utils.installer_cli import main_installer_cli as devops_devapps_install_main
                devops_devapps_install_main(group=True, which=maybe_a_group, interactive=False)
                console.print("✅ CLI applications installed successfully", style="bold green")
            except Exception as e:
                console.print(f"❌ Error installing CLI applications: {e}", style="bold red")
            if platform.system() != "Windows":
                run_shell_script(". $HOME/.bashrc")

    if "install_machineconfig" in selected_options:
        console.print(Panel("🐍 [bold green]PYTHON ENVIRONMENT[/bold green]\n[italic]Virtual environment setup[/italic]", border_style="green"))
        from machineconfig.scripts.python.helpers.helpers_devops.cli_self import install
        install(copy_assets=True, dev=False)

    if "install_ssh_server" in selected_options:
        console.print(Panel("🔒 [bold red]SSH SERVER[/bold red]\n[italic]Remote access setup[/italic]", border_style="red"))
        if platform.system() == "Windows":
            powershell_script = """Write-Host "🔧 Installing and configuring SSH server..."
Add-WindowsCapability -Online -Name OpenSSH.Server~~~~0.0.1.0
Start-Service sshd
Set-Service -Name sshd -StartupType 'Automatic'"""
            run_shell_script(f'powershell -Command "{powershell_script}"')
        else:
            run_shell_script("sudo nala install openssh-server -y")

    if "install_shell_profile" in selected_options:
        console.print(Panel("🐚 [bold green]SHELL PROFILE[/bold green]\n[italic]Shell configuration setup[/italic]", border_style="green"))
        console.print("🔧 Configuring shell profile", style="bold cyan")
        try:
            from machineconfig.profile.create_shell_profile import create_default_shell_profile
            create_default_shell_profile()
            console.print("✅ Shell profile configured successfully", style="bold green")
            from machineconfig.profile.create_links_export import main_public_from_parser
            main_public_from_parser(method="copy", on_conflict="overwrite-default-path", which="all", interactive=False)
            if platform.system() == "Windows":
                from machineconfig.jobs.installer.python_scripts.nerfont_windows_helper import install_nerd_fonts
                install_nerd_fonts()
                from machineconfig.settings.wt.set_wt_settings import main as set_wt_settings_main
                set_wt_settings_main()
        except Exception as e:
            console.print(f"❌ Error configuring shell profile: {e}", style="bold red")

    if "retrieve_repositories" in selected_options:
        console.print(Panel("📚 [bold bright_magenta]REPOSITORIES[/bold bright_magenta]\n[italic]Project code retrieval[/italic]", border_style="bright_magenta"))
        from machineconfig.scripts.python.helpers.helpers_devops import cli_repos
        cli_repos.clone(directory=str(Path.home() / "code"), cloud=None)

    if "retrieve_data" in selected_options:
        console.print(Panel("💾 [bold bright_cyan]DATA RETRIEVAL[/bold bright_cyan]\n[italic]Backup restoration[/italic]", border_style="bright_cyan"))
        console.print("🔧 Retrieving backup data", style="bold cyan")
        try:
            from machineconfig.scripts.python.helpers.helpers_devops.devops_backup_retrieve import main_backup_retrieve
            main_backup_retrieve(direction="RETRIEVE", cloud=None, which=None)
            console.print("✅ Backup data retrieved successfully", style="bold green")
        except Exception as e:
            console.print(f"❌ Error retrieving backup data: {e}", style="bold red")


def main() -> None:
    display_header()
    selected_options = get_installation_choices()
    if not selected_options:
        console.print("❌ No options selected. Exiting...", style="bold red")
        sys.exit(0)
    console.print(f"\n✅ Selected options: {'\n'.join(selected_options)}", style="bold green")
    proceed = questionary.confirm("🚀 Proceed with installation?", default=True).ask()
    if not proceed:
        console.print("❌ Installation cancelled.", style="bold red")
        sys.exit(0)
    execute_installations(selected_options=selected_options)
    display_completion_message()


if __name__ == "__main__":
    pass



from machineconfig.utils.source_of_truth import EXCLUDE_DIRS
from rich.console import Console
from rich.panel import Panel
import platform
import subprocess
from pathlib import Path
from typing import Optional

console = Console()


def sanitize_path(a_path: str) -> Path:
    path = Path(a_path)
    if Path.cwd() == Path.home() and not path.exists():
        result = input("Current working directory is home, and passed path is not full path, are you sure you want to continue, [y]/n? ") or "y"
        if result == "y":
            import sys

            sys.exit()
    if path.as_posix().startswith("/home") or path.as_posix().startswith("/Users"):
        if platform.system() == "Windows":  # path copied from Linux/Mac to Windows
            # For Linux: /home/username, for Mac: /Users/username
            skip_parts = 3 if path.as_posix().startswith("/home") else 3  # Both have 3 parts to skip
            path = Path.home().joinpath(*path.parts[skip_parts:])
            assert path.exists(), f"File not found: {path}"
            source_os = "Linux" if path.as_posix().startswith("/home") else "macOS"
            console.print(Panel(f"🔗 PATH MAPPING | {source_os} → Windows: `{a_path}` ➡️ `{path}`", title="Path Mapping", expand=False))
        elif platform.system() in ["Linux", "Darwin"] and Path.home().as_posix() not in path.as_posix():  # copied between Unix-like systems with different username
            skip_parts = 3  # Both /home/username and /Users/username have 3 parts to skip
            path = Path.home().joinpath(*path.parts[skip_parts:])
            assert path.exists(), f"File not found: {path}"
            current_os = "Linux" if platform.system() == "Linux" else "macOS"
            source_os = "Linux" if path.as_posix().startswith("/home") else "macOS"
            console.print(Panel(f"🔗 PATH MAPPING | {source_os} → {current_os}: `{a_path}` ➡️ `{path}`", title="Path Mapping", expand=False))
    elif path.as_posix().startswith("C:"):
        if platform.system() in ["Linux", "Darwin"]:  # path copied from Windows to Linux/Mac
            xx = str(a_path).replace("\\\\", "/")
            path = Path.home().joinpath(*Path(xx).parts[3:])  # exclude C:\\Users\\username
            assert path.exists(), f"File not found: {path}"
            target_os = "Linux" if platform.system() == "Linux" else "macOS"
            console.print(Panel(f"🔗 PATH MAPPING | Windows → {target_os}: `{a_path}` ➡️ `{path}`", title="Path Mapping", expand=False))
        elif platform.system() == "Windows" and Path.home().as_posix() not in path.as_posix():  # copied from Windows to Windows with different username
            path = Path.home().joinpath(*path.parts[2:])
            assert path.exists(), f"File not found: {path}"
            console.print(Panel(f"🔗 PATH MAPPING | Windows → Windows: `{a_path}` ➡️ `{path}`", title="Path Mapping", expand=False))
    return path.expanduser().absolute()


def find_scripts(root: Path, name_substring: str, suffixes: set[str]) -> tuple[list[Path], list[Path]]:
    filename_matches = []
    partial_path_matches = []
    for entry in root.iterdir():
        if entry.is_dir():
            if entry.name in set(EXCLUDE_DIRS):
                # prune this entire subtree
                continue
            tmp1, tmp2 = find_scripts(entry, name_substring, suffixes)
            filename_matches.extend(tmp1)
            partial_path_matches.extend(tmp2)
        elif entry.is_file() and entry.suffix in suffixes:
            if name_substring.lower() in entry.name.lower():
                filename_matches.append(entry)
            elif name_substring.lower() in entry.as_posix().lower():
                partial_path_matches.append(entry)
    return filename_matches, partial_path_matches


def match_file_name(sub_string: str, search_root: Path, suffixes: set[str]) -> Path:
    search_root_obj = search_root.absolute()
    # assume subscript is filename only, not a sub_path. There is no need to fzf over the paths.
    filename_matches, partial_path_matches = find_scripts(search_root_obj, sub_string, suffixes)
    if len(filename_matches) == 1:
        return Path(filename_matches[0])
    console.print(Panel(f"Partial filename {search_root_obj} match with case-insensitivity failed. This generated #{len(filename_matches)} results.", title="Search", expand=False))
    if len(filename_matches) < 20:
        print("\n".join([a_potential_match.as_posix() for a_potential_match in filename_matches]))
    if len(filename_matches) > 1:
        print(f"Try to narrow down filename_matches search by case-sensitivity, found {len(filename_matches)} results. First @ {filename_matches[0].as_posix()}")
        # let's see if avoiding .lower() helps narrowing down to one result
        reduced_scripts = [a_potential_match for a_potential_match in filename_matches if sub_string in a_potential_match.name]
        if len(reduced_scripts) == 1:
            return Path(reduced_scripts[0])
        elif len(reduced_scripts) > 1:
            from machineconfig.utils.options import choose_from_options
            choice = choose_from_options(multi=False, msg="Multiple matches found", options=reduced_scripts, tv=True)
            return Path(choice)
        print(f"Result: This still generated {len(reduced_scripts)} results.")
        if len(reduced_scripts) < 10:
            print("\n".join([a_potential_match.as_posix() for a_potential_match in reduced_scripts]))

    console.print(Panel(f"Partial path match with case-insensitivity failed. This generated #{len(partial_path_matches)} results.", title="Search", expand=False))
    if len(partial_path_matches) == 1:
        return Path(partial_path_matches[0])
    elif len(partial_path_matches) > 1:
        print("Try to narrow down partial_path_matches search by case-sensitivity.")
        reduced_scripts = [a_potential_match for a_potential_match in partial_path_matches if sub_string in a_potential_match.as_posix()]
        if len(reduced_scripts) == 1:
            return Path(reduced_scripts[0])
        print(f"Result: This still generated {len(reduced_scripts)} results.")        

    try:

        if len(partial_path_matches) == 0:
            print("No partial path matches found, trying to do fd with --no-ignore ...")
            fzf_cmd = f"cd '{search_root_obj}'; fd --no-ignore --type file --strip-cwd-prefix | fzf --ignore-case --exact --query={sub_string}"
        else:
            fzf_cmd = f"cd '{search_root_obj}'; fd --type file --strip-cwd-prefix | fzf --ignore-case --exact --query={sub_string}"
        console.print(Panel(f"🔍 Second attempt: SEARCH STRATEGY | Using fd to search for '{sub_string}' in '{search_root_obj}' ...\n{fzf_cmd}", title="Search Strategy", expand=False))
        search_res_raw = subprocess.run(fzf_cmd, stdout=subprocess.PIPE, text=True, check=True, shell=True).stdout
        search_res = search_res_raw.strip().split("\n")
    except subprocess.CalledProcessError as cpe:
        console.print(Panel(f"❌ ERROR | FZF search failed with '{sub_string}' in '{search_root_obj}'.\n{cpe}", title="Error", expand=False))
        import sys
        sys.exit(f"💥 FILE NOT FOUND | Path {sub_string} does not exist @ root {search_root_obj}. No search results.")
    if len(search_res) == 1:
        return search_root_obj.joinpath(search_res_raw)
    elif len(search_res) == 0:
        msg = Panel(f"💥 FILE NOT FOUND | Path {sub_string} does not exist @ root {search_root_obj}. No search results", title="File Not Found", expand=False)
        raise FileNotFoundError(msg)

    print(f"⚠️ WARNING | Multiple search results found for `{sub_string}`:\n'{search_res}'")
    cmd = f"cd '{search_root_obj}'; fd --type file | fzf --select-1 --query={sub_string}"
    console.print(Panel(f"🔍 SEARCH STRATEGY | Trying with raw fzf search ...\n{cmd}", title="Search Strategy", expand=False))
    try:
        res = subprocess.run(cmd, check=True, stdout=subprocess.PIPE, text=True, shell=True).stdout.strip()
    except subprocess.CalledProcessError as cpe:
        console.print(Panel(f"❌ ERROR | FZF search failed with '{sub_string}' in '{search_root_obj}'. {cpe}", title="Error", expand=False))
        msg = Panel(f"💥 FILE NOT FOUND | Path {sub_string} does not exist @ root {search_root_obj}. No search results", title="File Not Found", expand=False)
        raise FileNotFoundError(msg) from cpe
    return search_root_obj.joinpath(res)


def search_for_files_of_interest(path_obj: Path, suffixes: set[str]) -> list[Path]:
    if path_obj.is_file():
        return [path_obj]
    files: list[Path] = []
    directories_to_visit: list[Path] = [path_obj]
    while directories_to_visit:
        current_dir = directories_to_visit.pop()
        for entry in current_dir.iterdir():
            if entry.is_dir():
                if entry.name == ".venv":
                    continue
                directories_to_visit.append(entry)
                continue
            if entry.suffix not in suffixes:
                continue
            if entry.suffix == ".py" and entry.name == "__init__.py":
                continue
            files.append(entry)
    return files


def get_choice_file(path: str, suffixes: Optional[set[str]]):
    path_obj = sanitize_path(path)
    if suffixes is None:
        import platform
        if platform.system() == "Windows":
            suffixes = {".py", ".ps1", ".sh"}
        elif platform.system() in ["Linux", "Darwin"]:
            suffixes = {".py", ".sh"}
        else:
            suffixes = {".py"}
    if not path_obj.exists():
        print(f"🔍 Searching for file matching `{path}` under `{Path.cwd()}`, but only if suffix matches {suffixes}")
        choice_file = match_file_name(sub_string=path, search_root=Path.cwd(), suffixes=suffixes)
    elif path_obj.is_dir():
        print(f"🔍 Searching recursively for Python, PowerShell and Shell scripts in directory `{path_obj}`")
        files = search_for_files_of_interest(path_obj, suffixes=suffixes)
        print(f"🔍 Got #{len(files)} results.")
        from machineconfig.utils.options import choose_from_options
        choice_file = choose_from_options(multi=False, options=files, tv=True, msg="Choose one option")
        choice_file = Path(choice_file)
    else:
        choice_file = path_obj
    return choice_file



"""devops with emojis - lazy loading subcommands."""

import typer
from typing import Optional, Annotated, Literal


def install(which: Annotated[Optional[str], typer.Argument(..., help="Comma-separated list of program names to install, or group name if --group flag is set.")] = None,
        group: Annotated[bool, typer.Option(..., "--group", "-g", help="Treat 'which' as a group name. A group is bundle of apps.")] = False,
        interactive: Annotated[bool, typer.Option(..., "--interactive", "-i", help="Interactive selection of programs to install.")] = False,
    ) -> None:
        """📦 Install packages"""
        import machineconfig.utils.installer_utils.installer_cli as installer_entry_point
        installer_entry_point.main_installer_cli(which=which, group=group, interactive=interactive)


def repos(ctx: typer.Context) -> None:
    """📁 [r] Manage development repositories"""
    import machineconfig.scripts.python.helpers.helpers_devops.cli_repos as cli_repos
    cli_repos.get_app()(ctx.args, standalone_mode=False)


def config(ctx: typer.Context) -> None:
    """⚙️ [c] Configuration management"""
    import machineconfig.scripts.python.helpers.helpers_devops.cli_config as cli_config
    cli_config.get_app()(ctx.args, standalone_mode=False)


def data(ctx: typer.Context) -> None:
    """💾 [d] Data management"""
    import machineconfig.scripts.python.helpers.helpers_devops.cli_data as cli_data
    cli_data.get_app()(ctx.args, standalone_mode=False)


def self_cmd(ctx: typer.Context) -> None:
    """🔧 [s] Self management"""
    import machineconfig.scripts.python.helpers.helpers_devops.cli_self as cli_self
    cli_self.get_app()(ctx.args, standalone_mode=False)


def network(ctx: typer.Context) -> None:
    """🌐 [n] Network management"""
    import machineconfig.scripts.python.helpers.helpers_devops.cli_nw as cli_network
    cli_network.get_app()(ctx.args, standalone_mode=False)


def execute(
    name: Annotated[str, typer.Argument(help="Name of script to run, e.g., 'a' for a.py, or command to execute")] = "",
    where: Annotated[Literal["all", "a", "private", "p", "public", "b", "library", "l", "dynamic", "d", "custom", "c"], typer.Option("--where", "-w", help="Where to look for the script")] = "all",
    interactive: Annotated[bool, typer.Option(..., "--interactive", "-i", help="Interactive selection of scripts to run")] = False,
    command: Annotated[Optional[bool], typer.Option(..., "--command", "-c", help="Run as command")] = False,
    list_scripts: Annotated[bool, typer.Option(..., "--list", "-l", help="List available scripts in all locations")] = False,
) -> None:
    """▶️ Execute python/shell scripts from pre-defined directories or as command."""
    import machineconfig.scripts.python.helpers.helpers_devops.run_script as run_py_script_module
    run_py_script_module.run_py_script(name=name, where=where, interactive=interactive, command=command, list_scripts=list_scripts)


def get_app() -> typer.Typer:
    cli_app = typer.Typer(help="🛠️ DevOps operations", no_args_is_help=True, add_help_option=True, add_completion=False)
    ctx_settings: dict[str, object] = {"allow_extra_args": True, "allow_interspersed_args": True, "ignore_unknown_options": True, "help_option_names": []}

    cli_app.command("install", no_args_is_help=True, help=install.__doc__, short_help="🛠️ [i] Install essential packages")(install)
    cli_app.command("i", no_args_is_help=True, help=install.__doc__, hidden=True)(install)

    cli_app.command("repos", help="📁 [r] Manage development repositories", context_settings=ctx_settings)(repos)
    cli_app.command("r", hidden=True, context_settings=ctx_settings)(repos)
    cli_app.command("config", help="⚙️ [c] Configuration management", context_settings=ctx_settings)(config)
    cli_app.command("c", hidden=True, context_settings=ctx_settings)(config)
    cli_app.command("data", help="💾 [d] Data management", context_settings=ctx_settings)(data)
    cli_app.command("d", hidden=True, context_settings=ctx_settings)(data)
    cli_app.command("self", help="🔧 [s] Self management", context_settings=ctx_settings)(self_cmd)
    cli_app.command("s", hidden=True, context_settings=ctx_settings)(self_cmd)
    cli_app.command("network", help="🌐 [n] Network management", context_settings=ctx_settings)(network)
    cli_app.command("n", hidden=True, context_settings=ctx_settings)(network)

    cli_app.command("execute", no_args_is_help=True, short_help="▶️ [e] Execute python/shell scripts from pre-defined directories or as command")(execute)
    cli_app.command("e", no_args_is_help=True, hidden=True)(execute)

    return cli_app


def main():
    app = get_app()
    app()



"""fire - Fire and manage jobs."""

from typing import Optional, Annotated
import typer


def fire(
    ctx: typer.Context,
    path: Annotated[str, typer.Argument(help="Path to the Python file to run")] = ".",
    function: Annotated[Optional[str], typer.Argument(help="Function to run")] = None,
    ve: Annotated[str, typer.Option("--ve", "-v", help="Virtual environment name")] = "",
    cmd: Annotated[bool, typer.Option("--cmd", "-B", help="Create a cmd fire command to launch the job asynchronously")] = False,
    interactive: Annotated[bool, typer.Option("--interactive", "-i", help="Whether to run the job interactively using IPython")] = False,
    debug: Annotated[bool, typer.Option("--debug", "-d", help="Enable debug mode")] = False,
    choose_function: Annotated[bool, typer.Option("--choose-function", "-c", help="Choose function interactively")] = False,
    loop: Annotated[bool, typer.Option("--loop", "-l", help="Infinite recursion (runs again after completion/interruption)")] = False,
    jupyter: Annotated[bool, typer.Option("--jupyter", "-j", help="Open in a jupyter notebook")] = False,
    marimo: Annotated[bool, typer.Option("--marimo", "-M", help="Open in a marimo notebook")] = False,
    module: Annotated[bool, typer.Option("--module", "-m", help="Launch the main file")] = False,
    script: Annotated[bool, typer.Option("--script", "-s", help="Launch as a script without fire")] = False,
    optimized: Annotated[bool, typer.Option("--optimized", "-O", help="Run the optimized version of the function")] = False,
    zellij_tab: Annotated[Optional[str], typer.Option("--zellij-tab", "-z", help="Open in a new zellij tab")] = None,
    submit_to_cloud: Annotated[bool, typer.Option("--submit-to-cloud", "-C", help="Submit to cloud compute")] = False,
    remote: Annotated[bool, typer.Option("--remote", "-r", help="Launch on a remote machine")] = False,
    streamlit: Annotated[bool, typer.Option("--streamlit", "-S", help="Run as streamlit app")] = False,
    environment: Annotated[str, typer.Option("--environment", "-E", help="Choose ip, localhost, hostname or arbitrary url")] = "",
    holdDirectory: Annotated[bool, typer.Option("--holdDirectory", "-D", help="Hold current directory and avoid cd'ing to the script directory")] = False,
    PathExport: Annotated[bool, typer.Option("--PathExport", "-P", help="Augment the PYTHONPATH with repo root")] = False,
    git_pull: Annotated[bool, typer.Option("--git-pull", "-g", help="Start by pulling the git repo")] = False,
    watch: Annotated[bool, typer.Option("--watch", "-w", help="Watch the file for changes")] = False,
) -> None:
    """Main function to process fire jobs arguments."""
    from machineconfig.scripts.python.helpers.helpers_fire_command.fire_jobs_args_helper import FireJobArgs, parse_fire_args_from_context
    from machineconfig.scripts.python.helpers.helpers_fire_command.fire_jobs_impl import route

    fire_args = parse_fire_args_from_context(ctx)

    args = FireJobArgs(
        path=path,
        function=function,
        ve=ve,
        cmd=cmd,
        interactive=interactive,
        debug=debug,
        choose_function=choose_function,
        loop=loop,
        jupyter=jupyter,
        marimo=marimo,
        submit_to_cloud=submit_to_cloud,
        remote=remote,
        module=module,
        script=script,
        streamlit=streamlit,
        environment=environment,
        holdDirectory=holdDirectory,
        PathExport=PathExport,
        git_pull=git_pull,
        optimized=optimized,
        zellij_tab=zellij_tab,
        watch=watch,
    )
    try:
        route(args, fire_args)
    except SystemExit:
        raise
    except Exception as e:
        import sys
        print(f"❌ Error: {e}", file=sys.stderr)
        sys.exit(1)


def get_app() -> typer.Typer:
    app = typer.Typer(add_completion=False)
    app.command(context_settings={"allow_extra_args": True, "allow_interspersed_args": False})(fire)
    return app


def main() -> None:
    app = get_app()
    app()


if __name__ == "__main__":
    pass





from typing import Optional, Union
from pathlib import Path
from machineconfig.utils.accessories import randstr
from machineconfig.utils.meta import lambda_to_python_script
from machineconfig.utils.ssh_utils.abc import MACHINECONFIG_VERSION, DEFAULT_PICKLE_SUBDIR


def copy_to_here(
    self: "SSH",
    source: Union[str, Path],
    target: Optional[Union[str, Path]],
    compress_with_zip: bool = False,
    recursive: bool = False,
    internal_call: bool = False,
) -> None:
    if self.sftp is None:
        raise RuntimeError(f"SFTP connection not available for {self.hostname}. Cannot transfer files.")

    if not internal_call:
        print(f"{'⬇️' * 5} SFTP DOWNLOADING FROM `{source}` TO `{target}`")

    source_obj = Path(source)
    expanded_source = self.expand_remote_path(source_path=source_obj)

    if not compress_with_zip:
        is_dir = self.check_remote_is_dir(source_path=expanded_source)

        if is_dir:
            if not recursive:
                raise RuntimeError(
                    f"SSH Error: source `{source_obj}` is a directory! Set recursive=True for recursive transfer or compress_with_zip=True to zip it."
                )

            def search_files(directory_path: str, json_output_path: str) -> list[str]:
                from pathlib import Path
                import json

                file_paths_list = [
                    file_path.as_posix() for file_path in Path(directory_path).expanduser().absolute().rglob("*") if file_path.is_file()
                ]
                json_result_path = Path(json_output_path)
                json_result_path.parent.mkdir(parents=True, exist_ok=True)
                json_result_path.write_text(json.dumps(file_paths_list, indent=2), encoding="utf-8")
                print(json_result_path.as_posix())
                return file_paths_list

            remote_json_output = Path.home().joinpath(f"{DEFAULT_PICKLE_SUBDIR}/return_{randstr()}.json").as_posix()
            command = lambda_to_python_script(
                lambda: search_files(directory_path=expanded_source, json_output_path=remote_json_output),
        in_global=True, import_module=False
            )
            response = self.run_py_remotely(
                python_code=command,
                uv_with=[MACHINECONFIG_VERSION],
                uv_project_dir=None,
                description="Searching for files in source",
                verbose_output=False,
                strict_stderr=False,
                strict_return_code=False,
            )
            remote_json_path = response.op.strip()
            if not remote_json_path:
                raise RuntimeError(f"Could not resolve source path {source} - no response from remote")

            local_json = Path.home().joinpath(f"{DEFAULT_PICKLE_SUBDIR}/local_{randstr()}.json")
            self.simple_sftp_get(remote_path=remote_json_path, local_path=local_json)
            import json

            try:
                source_list_str = json.loads(local_json.read_text(encoding="utf-8"))
            except (json.JSONDecodeError, FileNotFoundError) as err:
                raise RuntimeError(f"Could not resolve source path {source} - invalid JSON response: {err}") from err
            finally:
                if local_json.exists():
                    local_json.unlink()
            assert isinstance(source_list_str, list), f"Could not resolve source path {source}"
            file_paths_to_download = [Path(file_path_str) for file_path_str in source_list_str]

            if target is None:

                def collapse_to_home_dir(absolute_path: str, json_output_path: str) -> str:
                    from pathlib import Path
                    import json

                    source_absolute_path = Path(absolute_path).expanduser().absolute()
                    try:
                        relative_to_home = source_absolute_path.relative_to(Path.home())
                        collapsed_path_posix = (Path("~") / relative_to_home).as_posix()
                        json_result_path = Path(json_output_path)
                        json_result_path.parent.mkdir(parents=True, exist_ok=True)
                        json_result_path.write_text(json.dumps(collapsed_path_posix, indent=2), encoding="utf-8")
                        print(json_result_path.as_posix())
                        return collapsed_path_posix
                    except ValueError:
                        raise RuntimeError(f"Source path must be relative to home directory: {source_absolute_path}")

                remote_json_output = Path.home().joinpath(f"{DEFAULT_PICKLE_SUBDIR}/return_{randstr()}.json").as_posix()
                command = lambda_to_python_script(
                    lambda: collapse_to_home_dir(absolute_path=expanded_source, json_output_path=remote_json_output),
                    in_global=True,
                    import_module=False,
                )
                response = self.run_py_remotely(
                    python_code=command,
                    uv_with=[MACHINECONFIG_VERSION],
                    uv_project_dir=None,
                    description="Finding default target via relative source path",
                    verbose_output=False,
                    strict_stderr=False,
                    strict_return_code=False,
                )
                remote_json_path_dir = response.op.strip()
                if not remote_json_path_dir:
                    raise RuntimeError("Could not resolve target path - no response from remote")

                local_json_dir = Path.home().joinpath(f"{DEFAULT_PICKLE_SUBDIR}/local_{randstr()}.json")
                self.simple_sftp_get(remote_path=remote_json_path_dir, local_path=local_json_dir)
                import json

                try:
                    target_dir_str = json.loads(local_json_dir.read_text(encoding="utf-8"))
                except (json.JSONDecodeError, FileNotFoundError) as err:
                    raise RuntimeError(f"Could not resolve target path - invalid JSON response: {err}") from err
                finally:
                    if local_json_dir.exists():
                        local_json_dir.unlink()
                assert isinstance(target_dir_str, str), "Could not resolve target path"
                target = Path(target_dir_str)

            target_dir = Path(target).expanduser().absolute()

            for idx, file_path in enumerate(file_paths_to_download):
                print(f"   {idx + 1:03d}. {file_path}")

            for file_path in file_paths_to_download:
                local_file_target = target_dir.joinpath(Path(file_path).relative_to(expanded_source))
                self.copy_to_here(source=file_path, target=local_file_target, compress_with_zip=False, recursive=False, internal_call=True)

            return None

    if compress_with_zip:
        print("🗜️ ZIPPING ...")

        def zip_source(path_to_zip: str, json_output_path: str) -> str:
            from pathlib import Path
            import shutil
            import json

            source_to_compress = Path(path_to_zip).expanduser().absolute()
            archive_base_path = source_to_compress.parent / (source_to_compress.name + "_archive")
            if source_to_compress.is_dir():
                shutil.make_archive(str(archive_base_path), "zip", source_to_compress)
            else:
                shutil.make_archive(str(archive_base_path), "zip", source_to_compress.parent, source_to_compress.name)
            zip_file_path = str(archive_base_path) + ".zip"
            json_result_path = Path(json_output_path)
            json_result_path.parent.mkdir(parents=True, exist_ok=True)
            json_result_path.write_text(json.dumps(zip_file_path, indent=2), encoding="utf-8")
            print(json_result_path.as_posix())
            return zip_file_path

        remote_json_output = Path.home().joinpath(f"{DEFAULT_PICKLE_SUBDIR}/return_{randstr()}.json").as_posix()
        command = lambda_to_python_script(
            lambda: zip_source(path_to_zip=expanded_source, json_output_path=remote_json_output),
        in_global=True, import_module=False
        )
        response = self.run_py_remotely(
            python_code=command,
            uv_with=[MACHINECONFIG_VERSION],
            uv_project_dir=None,
            description=f"Zipping source file {source}",
            verbose_output=False,
            strict_stderr=False,
            strict_return_code=False,
        )
        remote_json_path = response.op.strip()
        if not remote_json_path:
            raise RuntimeError(f"Could not zip {source} - no response from remote")

        local_json = Path.home().joinpath(f"{DEFAULT_PICKLE_SUBDIR}/local_{randstr()}.json")
        self.simple_sftp_get(remote_path=remote_json_path, local_path=local_json)
        import json

        try:
            zipped_path = json.loads(local_json.read_text(encoding="utf-8"))
        except (json.JSONDecodeError, FileNotFoundError) as err:
            raise RuntimeError(f"Could not zip {source} - invalid JSON response: {err}") from err
        finally:
            if local_json.exists():
                local_json.unlink()
        assert isinstance(zipped_path, str), f"Could not zip {source}"
        source_obj = Path(zipped_path)
        expanded_source = zipped_path

    if target is None:

        def collapse_to_home(absolute_path: str, json_output_path: str) -> str:
            from pathlib import Path
            import json

            source_absolute_path = Path(absolute_path).expanduser().absolute()
            try:
                relative_to_home = source_absolute_path.relative_to(Path.home())
                collapsed_path_posix = (Path("~") / relative_to_home).as_posix()
                json_result_path = Path(json_output_path)
                json_result_path.parent.mkdir(parents=True, exist_ok=True)
                json_result_path.write_text(json.dumps(collapsed_path_posix, indent=2), encoding="utf-8")
                print(json_result_path.as_posix())
                return collapsed_path_posix
            except ValueError:
                raise RuntimeError(f"Source path must be relative to home directory: {source_absolute_path}")

        remote_json_output = Path.home().joinpath(f"{DEFAULT_PICKLE_SUBDIR}/return_{randstr()}.json").as_posix()
        command = lambda_to_python_script(
            lambda: collapse_to_home(absolute_path=expanded_source, json_output_path=remote_json_output),
        in_global=True, import_module=False
        )
        response = self.run_py_remotely(
            python_code=command,
            uv_with=[MACHINECONFIG_VERSION],
            uv_project_dir=None,
            description="Finding default target via relative source path",
            verbose_output=False,
            strict_stderr=False,
            strict_return_code=False,
        )
        remote_json_path = response.op.strip()
        if not remote_json_path:
            raise RuntimeError("Could not resolve target path - no response from remote")

        local_json = Path.home().joinpath(f"{DEFAULT_PICKLE_SUBDIR}/local_{randstr()}.json")
        self.simple_sftp_get(remote_path=remote_json_path, local_path=local_json)
        import json

        try:
            target_str = json.loads(local_json.read_text(encoding="utf-8"))
        except (json.JSONDecodeError, FileNotFoundError) as err:
            raise RuntimeError(f"Could not resolve target path - invalid JSON response: {err}") from err
        finally:
            if local_json.exists():
                local_json.unlink()
        assert isinstance(target_str, str), "Could not resolve target path"
        target = Path(target_str)
        assert str(target).startswith("~"), f"If target is not specified, source must be relative to home.\n{target=}"

    target_obj = Path(target).expanduser().absolute()
    target_obj.parent.mkdir(parents=True, exist_ok=True)

    if compress_with_zip and target_obj.suffix != ".zip":
        target_obj = target_obj.with_suffix(target_obj.suffix + ".zip")

    print(f"""📥 [DOWNLOAD] Receiving: {expanded_source}  ==>  Local Path: {target_obj}""")
    try:
        with self.tqdm_wrap(ascii=True, unit="b", unit_scale=True) as pbar:
            if self.sftp is None:  # type: ignore[unreachable]
                raise RuntimeError(f"SFTP connection lost for {self.hostname}")
            self.sftp.get(remotepath=expanded_source, localpath=str(target_obj), callback=pbar.view_bar)  # type: ignore
    except Exception:
        if target_obj.exists():
            target_obj.unlink()
        raise

    if compress_with_zip:
        import zipfile

        extract_to = target_obj.parent / target_obj.stem
        with zipfile.ZipFile(target_obj, "r") as zip_ref:
            zip_ref.extractall(extract_to)
        target_obj.unlink()
        target_obj = extract_to

        def delete_temp_zip(path_to_delete: str) -> None:
            from pathlib import Path
            import shutil

            file_or_dir_path = Path(path_to_delete)
            if file_or_dir_path.exists():
                if file_or_dir_path.is_dir():
                    shutil.rmtree(file_or_dir_path)
                else:
                    file_or_dir_path.unlink()

        command = lambda_to_python_script(lambda: delete_temp_zip(path_to_delete=expanded_source),
                                in_global=True, import_module=False)
        self.run_py_remotely(
            python_code=command,
            uv_with=[MACHINECONFIG_VERSION],
            uv_project_dir=None,
            description="Cleaning temp zip files @ remote.",
            verbose_output=False,
            strict_stderr=True,
            strict_return_code=True,
        )

    print("\n")
    return None


if __name__ == "__main__":
    from machineconfig.utils.ssh import SSH




import re
import subprocess
from pathlib import Path


def strip_ansi_codes(text: str) -> str:
    return re.sub(r'\x1b\[[0-9;]*[a-zA-Z]', '', text)


def choose_zellij_session(name: str | None, new_session: bool, kill_all: bool) -> tuple[str, str | None]:
    """Choose a Zellij session. Returns tuple of (action, script_to_run) where action is 'run_script', 'exit', or 'error'."""
    if name is not None:
        return ("run_script", f"zellij attach {name}")
    if new_session:
        cmd = "zellij --layout st2"
        if kill_all:
            cmd = f"zellij kill-sessions\n{cmd}"
        return ("run_script", cmd)
    cmd = "zellij list-sessions"
    sessions: list[str] = subprocess.check_output(cmd, shell=True).decode().strip().split("\n")
    sessions = [s for s in sessions if s.strip()]
    # print(f"Found Zellij sessions: {sessions}")
    sessions.sort(key=lambda s: "EXITED" in s)
    if "current" in sessions:
        return ("error", "Already in a Zellij session, avoiding nesting and exiting.")
    if len(sessions) == 0:
        return ("run_script", "zellij --layout st2")
    if len(sessions) == 1:
        sn = strip_ansi_codes(sessions[0])
        session_name = sn.split(" [Created")[0]
        return ("run_script", f"zellij attach {session_name}")
    from machineconfig.utils.options import choose_from_options
    NEW_SESSION_LABEL = "NEW SESSION"
    options = [NEW_SESSION_LABEL] + sessions
    session_name = choose_from_options(msg="Choose a Zellij session to attach to:", multi=False, options=options, tv=True)
    if session_name == NEW_SESSION_LABEL:
        cmd = "zellij --layout st2"
        if kill_all:
            cmd = f"zellij kill-sessions\n{cmd}"
        return ("run_script", cmd)
    session_name = session_name.split(" [Created")[0]
    return ("run_script", f"zellij attach {session_name}")


def get_session_tabs() -> list[tuple[str, str]]:
    cmd = "zellij list-sessions"
    sessions: list[str] = subprocess.check_output(cmd, shell=True).decode().strip().split("\n")
    sessions = [strip_ansi_codes(s) for s in sessions]
    active_sessions = [s for s in sessions if "EXITED" not in s]
    result: list[tuple[str, str]] = []
    for session_line in active_sessions:
        session_name = session_line.split(" [Created")[0].strip()
        tab_cmd = f"zellij  --session {session_name} action query-tab-names"
        try:
            tabs: list[str] = subprocess.check_output(tab_cmd, shell=True).decode().strip().split("\n")
            for tab in tabs:
                if tab.strip():
                    result.append((session_name, tab.strip()))
        except subprocess.CalledProcessError:
            continue
    return result


def start_wt(layout_name: str) -> tuple[str, str | None]:
    """Start a Windows Terminal layout by name. Returns tuple of (status, message) where status is 'success' or 'error'."""
    import json
    from machineconfig.utils.schemas.layouts.layout_types import LayoutsFile
    from machineconfig.cluster.sessions_managers.wt_local import run_wt_layout
    layouts_file = Path.home().joinpath("dotfiles/machineconfig/layouts.json")
    if not layouts_file.exists():
        return ("error", f"❌ Layouts file not found: {layouts_file}")
    layouts_data: LayoutsFile = json.loads(layouts_file.read_text(encoding="utf-8"))
    chosen_layout = next((a_layout for a_layout in layouts_data["layouts"] if a_layout["layoutName"] == layout_name), None)
    if not chosen_layout:
        available_layouts = [a_layout["layoutName"] for a_layout in layouts_data["layouts"]]
        return ("error", f"❌ Layout '{layout_name}' not found in layouts file.\nAvailable layouts: {', '.join(available_layouts)}")
    run_wt_layout(layout_config=chosen_layout)
    return ("success", None)





from typing import Literal, Annotated
from pathlib import Path
import typer
import machineconfig.scripts.python.helpers.helpers_devops.cli_config_dotfile as dotfile_module
import machineconfig.profile.create_links_export as create_links_export


def configure_shell_profile(which: Annotated[Literal["default", "d", "nushell", "n"], typer.Option(..., "--which", "-w", help="Which shell profile to create/configure")]="default"):
    """🔗 Configure your shell profile."""
    from machineconfig.profile.create_shell_profile import create_default_shell_profile,  create_nu_shell_profile
    match which:
        case "nushell" | "n":
            create_nu_shell_profile()
            return
        case "default" | "d":
            create_default_shell_profile()
            return
    typer.echo(f"[red]Error:[/] Unknown shell profile type: {which}")



def pwsh_theme():
    """🔗 Select powershell prompt theme."""
    import machineconfig.scripts.python.helpers.helpers_devops.themes as themes
    file = Path(themes.__file__).parent / "choose_pwsh_theme.ps1"
    import subprocess
    subprocess.run(["pwsh", "-File", str(file)])


def starship_theme():
    """🔗 Select starship prompt theme."""
    import subprocess
    from machineconfig.utils.code import run_shell_script
    
    presets: list[str] = ["catppuccin-powerline", "pastel-powerline", "tokyo-night", "gruvbox-rainbow", "jetpack"]
    config_path: Path = Path.home() / ".config" / "starship.toml"
    
    typer.echo("\n🚀 Starship Theme Selector\n")
    for idx, preset in enumerate(presets, start=1):
        typer.echo(f"{idx}. {preset}")
    
    choice: str = typer.prompt("Select a preset")
    
    try:
        choice_idx: int = int(choice)
        if 1 <= choice_idx <= len(presets):
            selected_preset: str = presets[choice_idx - 1]
            typer.echo(f"\n✨ Applying {selected_preset}...")
            run_shell_script(f"""starship preset {selected_preset} -o {config_path}""")
            typer.echo("\n📋 Preview:")
            subprocess.run(["starship", "module", "all"], check=False)
            typer.echo(f"\n✅ {selected_preset} applied!")
        else:
            typer.echo("❌ Invalid selection")
    except ValueError:
        typer.echo("❌ Please enter a valid number")


def copy_assets(which: Annotated[Literal["scripts", "s", "settings", "t", "both", "b"], typer.Argument(..., help="Which assets to copy")]):
    """🔗 Copy asset files from library to machine."""
    import machineconfig.profile.create_helper as create_helper
    match which:
        case "both" | "b":
            create_helper.copy_assets_to_machine(which="scripts")
            create_helper.copy_assets_to_machine(which="settings")
            return
        case "scripts" | "s":
            create_helper.copy_assets_to_machine(which="scripts")
            return
        case "settings" | "t":
            create_helper.copy_assets_to_machine(which="settings")
            return
    typer.echo(f"[red]Error:[/] Unknown asset type: {which}")



def get_app():
    config_apps = typer.Typer(help="⚙️ [c] configuration subcommands", no_args_is_help=True, add_help_option=True, add_completion=False)
    config_apps.command("private", no_args_is_help=True, help="🔗 [v] Manage private configuration files.")(create_links_export.main_private_from_parser)
    config_apps.command("v", no_args_is_help=True, hidden=True)(create_links_export.main_private_from_parser)
    config_apps.command("public", no_args_is_help=True, help="🔗 [b] Manage public configuration files.")(create_links_export.main_public_from_parser)
    config_apps.command("b", no_args_is_help=True, help="Manage public configuration files.", hidden=True)(create_links_export.main_public_from_parser)
    config_apps.command("dotfile", no_args_is_help=True, help="🔗 [d] Manage dotfiles.")(dotfile_module.main)
    config_apps.command("d", no_args_is_help=True,  hidden=True)(dotfile_module.main)
    config_apps.command("shell", no_args_is_help=False, help="🔗 [s] Configure your shell profile.")(configure_shell_profile)
    config_apps.command("s", no_args_is_help=False, help="Configure your shell profile.", hidden=True)(configure_shell_profile)
    config_apps.command("starship-theme", no_args_is_help=False, help="🔗 [t] Select starship prompt theme.")(starship_theme)
    config_apps.command("t", no_args_is_help=False, help="Select starship prompt theme.", hidden=True)(starship_theme)
    config_apps.command("pwsh-theme", no_args_is_help=False, help="🔗 [T] Select powershell prompt theme.")(pwsh_theme)
    config_apps.command("T", no_args_is_help=False, help="Select powershell prompt theme.", hidden=True)(pwsh_theme)

    config_apps.command("copy-assets", no_args_is_help=True, help="🔗 [c] Copy asset files from library to machine.", hidden=False)(copy_assets)
    config_apps.command("c", no_args_is_help=True, help="Copy asset files from library to machine.", hidden=True)(copy_assets)


    return config_apps



"""
Check Installations
===================

This module scans installed applications using VirusTotal and generates a safety report.
It also provides functionality to download and install pre-checked applications.
"""

import csv
import platform
from datetime import datetime
from typing import Optional

from rich.console import Console
from rich.progress import BarColumn, Progress, SpinnerColumn, TextColumn

from machineconfig.jobs.installer.checks.install_utils import upload_app
from machineconfig.jobs.installer.checks.report_utils import AppData, generate_markdown_report
from machineconfig.jobs.installer.checks.vt_utils import get_vt_client, scan_file
from machineconfig.utils.installer_utils.installer_runner import get_installed_cli_apps
from machineconfig.utils.path_extended import PathExtended
from machineconfig.utils.source_of_truth import CONFIG_ROOT, INSTALL_VERSION_ROOT

# Constants
APP_SUMMARY_PATH = CONFIG_ROOT.joinpath(f"profile/records/{platform.system().lower()}/apps_summary_report.csv")

console = Console()

def main() -> None:
    """Main execution function."""
    console.rule("[bold blue]MachineConfig Installation Checker[/bold blue]")

    # 1. Gather Installed Apps
    with console.status("[bold green]Gathering installed applications...[/bold green]"):
        apps_paths = get_installed_cli_apps()
        # Filter and find versions
        # This part mimics the original logic but simplifies it
        # We assume apps_paths contains PathExtended objects
        
        # Find version files
        install_version_root_ext = PathExtended(INSTALL_VERSION_ROOT)
        version_files = list(install_version_root_ext.search()) if install_version_root_ext.exists() else []
        
        apps_to_scan: list[tuple[PathExtended, Optional[str]]] = []
        
        for app_path in apps_paths:
            # Try to find version
            version = None
            # Simple matching logic: check if any version file matches the app name
            # This is a heuristic from the original code
            matching_versions = [v for v in version_files if v.stem == app_path.stem]
            if matching_versions:
                version = matching_versions[0].read_text(encoding="utf-8").strip()
            
            apps_to_scan.append((app_path, version))

    console.print(f"[green]Found {len(apps_to_scan)} applications to check.[/green]")

    # 2. Scan Apps with VirusTotal
    app_data_list: list[AppData] = []
    
    try:
        client = get_vt_client()
        
        with Progress(
            SpinnerColumn(),
            TextColumn("[progress.description]{task.description}"),
            BarColumn(),
            TextColumn("[progress.percentage]{task.percentage:>3.0f}%"),
            console=console
        ) as progress:
            scan_task = progress.add_task("[cyan]Scanning apps...", total=len(apps_to_scan))
            
            for app_path, version in apps_to_scan:
                progress.update(scan_task, description=f"Scanning {app_path.name}...")
                
                positive_pct, _ = scan_file(app_path, client, progress, scan_task)
                
                # Upload if safe (optional, based on original code logic which uploaded everything?)
                # Original code uploaded everything.
                progress.update(scan_task, description=f"Uploading {app_path.name}...")
                app_url = upload_app(app_path) or ""
                
                app_data_list.append({
                    "app_name": app_path.stem,
                    "version": version,
                    "positive_pct": positive_pct,
                    "scan_time": datetime.now().strftime("%Y-%m-%d %H:%M"),
                    "app_path": app_path.collapseuser(strict=False).as_posix(),
                    "app_url": app_url
                })
                
                progress.advance(scan_task)

    except FileNotFoundError as e:
        console.print(f"[bold red]{e}[/bold red]")
        console.print("[yellow]Skipping scanning due to missing credentials.[/yellow]")
        # Fallback: just list apps without scan results
        for app_path, version in apps_to_scan:
             app_data_list.append({
                "app_name": app_path.stem,
                "version": version,
                "positive_pct": None,
                "scan_time": datetime.now().strftime("%Y-%m-%d %H:%M"),
                "app_path": app_path.collapseuser(strict=False).as_posix(),
                "app_url": ""
            })
    except Exception as e:
        console.print(f"[bold red]An unexpected error occurred during scanning: {e}[/bold red]")

    # 3. Generate Reports
    if app_data_list:
        # CSV Report
        APP_SUMMARY_PATH.parent.mkdir(parents=True, exist_ok=True)
        csv_path = APP_SUMMARY_PATH.with_suffix(".csv")
        
        with open(csv_path, 'w', newline='', encoding='utf-8') as csvfile:
            fieldnames = list(app_data_list[0].keys())
            writer = csv.DictWriter(csvfile, fieldnames=fieldnames)
            writer.writeheader()
            writer.writerows(app_data_list)
        
        console.print(f"[green]CSV report saved to: {csv_path}[/green]")

        # Markdown Report
        md_path = APP_SUMMARY_PATH.with_suffix(".md")
        generate_markdown_report(app_data_list, md_path)
        console.print(f"[green]Markdown report saved to: {md_path}[/green]")



if __name__ == '__main__':
    main()



"""msearch - Machineconfig search helper."""

import typer
from typing import Annotated


def machineconfig_search(
    directory: Annotated[str, typer.Option(..., "--directory", "-d", help="The directory to search")] = ".",
    ast: Annotated[bool, typer.Option(..., "--ast", "-a", help="The abstract syntax tree search/ tree sitter search of symbols")] = False,
    symantic: Annotated[bool, typer.Option(..., "--symantic", "-s", help="The symantic search of symbols")] = False,
    extension: Annotated[str, typer.Option(..., "--extension", "-e", help="File extension to filter by (e.g., .py, .js)")] = "",
    file: Annotated[bool, typer.Option(..., "--file", "-f", help="File search using fzf")] = False,
    no_dotfiles: Annotated[bool, typer.Option(..., "--no-dotfiles", "-D", help="Exclude dotfiles from search")] = False,
    rga: Annotated[bool, typer.Option(..., "--rga", "-A", help="Use ripgrep-all for searching all (non text files) instead of ripgrep")] = False,
    install_dependencies: Annotated[bool, typer.Option(..., "--install-req", "-p", help="Install required dependencies if missing")] = False,
) -> None:
    """Machineconfig search helper."""
    from machineconfig.scripts.python.helpers.helpers_msearch.msearch_impl import machineconfig_search as impl
    impl(directory=directory, ast=ast, symantic=symantic, extension=extension, file=file, no_dotfiles=no_dotfiles, rga=rga, install_dependencies=install_dependencies)


def main() -> None:
    app = typer.Typer(add_completion=False, no_args_is_help=True)
    app.command(name="msearch", help=machineconfig_search.__doc__, short_help="machineconfig search helper", no_args_is_help=False)(machineconfig_search)
    app()





import typer
from typing import Annotated, Optional


def merge_pdfs(
        pdfs: Annotated[list[str], typer.Argument(..., help="Paths to the PDF files to merge.")],
        output: Annotated[Optional[str], typer.Option("--output", "-o", help="Output merged PDF file path.")] = None,
        compress: Annotated[bool, typer.Option("--compress", "-c", help="Compress the output PDF.")] = False,
    ) -> None:
    def merge_pdfs_internal(pdfs: list[str], output: str | None, compress: bool) -> None:
        from pypdf import PdfReader, PdfWriter
        writer = PdfWriter()
        for pdf_path in pdfs:
            reader = PdfReader(pdf_path)
            for page in reader.pages:
                writer.add_page(page)
        output_path = output if output else "merged.pdf"
        if compress:
            try:
                for p in writer.pages:
                    try:
                        # PageObject.compress_content_streams exists in pypdf
                        p.compress_content_streams()
                    except Exception:
                        # best-effort: ignore per-page compression failures
                        continue
            except Exception:
                pass
            try:
                writer.compress_identical_objects()
            except Exception:
                # non-fatal if this fails
                pass
        writer.write(output_path)
        print(f"✅ Merged PDF saved to: {output_path}")
    from machineconfig.utils.meta import lambda_to_python_script
    code = lambda_to_python_script(lambda : merge_pdfs_internal(pdfs=pdfs, output=output, compress=compress),
                                   in_global=True, import_module=False)
    from machineconfig.utils.code import run_shell_script, get_uv_command_executing_python_script
    uv_command, _py_file = get_uv_command_executing_python_script(python_script=code, uv_with=["pypdf"], uv_project_dir=None)
    run_shell_script(uv_command)


def compress_pdf(
        pdf_input: Annotated[str, typer.Argument(..., help="Path to the input PDF file to compress.")],
        output: Annotated[Optional[str], typer.Option("--output", "-o", help="Output compressed PDF file path.")] = None,
        quality: Annotated[int, typer.Option("--quality", "-q", help="JPEG quality for image compression (0-100, 0=no change, 100=best).")] = 85,
        image_dpi: Annotated[int, typer.Option("--image-dpi", "-d", help="Target DPI for image resampling. If set, images above this DPI will be downsampled.")] = 0,
        # remove_images: Annotated[bool, typer.Option("--remove-images", "-r", help="Remove all images from the PDF.")] = False,
        compress_streams: Annotated[bool, typer.Option("--compress-streams", "-c", help="Compress uncompressed streams.")] = True,
        use_objstms: Annotated[bool, typer.Option("--object-streams", "-s", help="Use object streams for additional compression.")] = True,
    ) -> None:
    def compress_pdf_internal(pdf_input: str, output: str | None, quality: int, image_dpi: int, compress_streams: bool, use_objstms: bool) -> None:
        import pymupdf
        from pathlib import Path
        output_path = output if output else pdf_input.replace(".pdf", "_compressed.pdf")
        doc = pymupdf.open(pdf_input)
        try:
            # if remove_images:
            #     for page in doc:
            #         page.remove_images()
            if quality > 0 or image_dpi > 0:
                doc.rewrite_images(
                    dpi_threshold=image_dpi if image_dpi > 0 else None,
                    dpi_target=max(72, image_dpi - 10) if image_dpi > 72 else 72,
                    quality=quality,
                    lossy=True,
                    lossless=True,
                )
            doc.save(
                output_path,
                deflate=compress_streams,
                garbage=3,
                use_objstms=1 if use_objstms else 0,
            )
            input_size = Path(pdf_input).stat().st_size
            output_size = Path(output_path).stat().st_size
            ratio = (1 - output_size / input_size) * 100
            print(f"✅ Compressed PDF saved to: {output_path}")
            print(f"   Original: {input_size / 1024 / 1024:.2f} MB")
            print(f"   Compressed: {output_size / 1024 / 1024:.2f} MB")
            print(f"   Reduction: {ratio:.1f}%")
        finally:
            doc.close()
    from machineconfig.utils.meta import lambda_to_python_script
    code = lambda_to_python_script(
        lambda: compress_pdf_internal(pdf_input=pdf_input, output=output, quality=quality, image_dpi=image_dpi, compress_streams=compress_streams, use_objstms=use_objstms),
        in_global=True,
        import_module=False,
    )
    from machineconfig.utils.code import run_shell_script, get_uv_command_executing_python_script
    uv_command, _py_file = get_uv_command_executing_python_script(python_script=code, uv_with=["pymupdf"], uv_project_dir=None)
    run_shell_script(uv_command)




#!/usr/bin/env -S uv run --script
# /// script
# requires-python = ">=3.13"
# dependencies = [
#     "machineconfig>=8.28",
#     "textual",
#     "pyperclip",
# ]
# ///



import os
import platform
from collections.abc import Mapping
from typing import Final

from rich.text import Text
from textual import on
from textual.app import App, ComposeResult
from textual.binding import Binding
from textual.containers import Horizontal, Vertical
from textual.widgets import Footer, Header, Label, ListItem, ListView, Static


VALUE_PREVIEW_LIMIT: Final[int] = 4096
SUMMARY_LIMIT: Final[int] = 96


def truncate_text(text: str, limit: int) -> tuple[str, int]:
    length = len(text)
    if length <= limit:
        return text, 0
    return text[:limit], length - limit


def format_summary(env_key: str, env_value: str, limit: int) -> str:
    sanitized = env_value.replace("\n", "\\n").replace("\t", "\\t")
    preview, remainder = truncate_text(sanitized, limit)
    if preview == "":
        base = f"{env_key} = <empty>"
    else:
        base = f"{env_key} = {preview}"
    if remainder == 0:
        return base
    return f"{base}... (+{remainder} chars)"


def collect_environment(env: Mapping[str, str]) -> list[tuple[str, str]]:
    return sorted(env.items(), key=lambda pair: pair[0].lower())


class EnvListItem(ListItem):
    def __init__(self, env_key: str, summary: str) -> None:
        super().__init__(Label(summary))
        self._env_key = env_key

    def env_key(self) -> str:
        return self._env_key


class EnvValuePreview(Static):
    def __init__(self, *args, **kwargs) -> None:  # type: ignore[no-untyped-def]
        super().__init__(*args, **kwargs)
        self.border_title = "Environment Value"

    def show_value(self, env_key: str, env_value: str) -> None:
        preview, remainder = truncate_text(env_value, VALUE_PREVIEW_LIMIT)
        text = Text()
        text.append(f"{env_key}\n\n", style="bold cyan")
        if preview == "":
            text.append("<empty>", style="dim")
        else:
            text.append(preview)
        if remainder > 0:
            text.append(f"\n... truncated {remainder} characters", style="yellow")
        self.update(text)


class StatusBar(Static):
    def __init__(self, *args, **kwargs) -> None:  # type: ignore[no-untyped-def]
        super().__init__(*args, **kwargs)
        self.border_title = "Status"

    def show_message(self, message: str, level: str) -> None:
        palette = {
            "info": "cyan",
            "success": "green",
            "warning": "yellow",
            "error": "red",
        }
        color = palette.get(level, "white")
        self.update(f"[{color}]{message}[/{color}]")


class EnvExplorerApp(App[None]):
    CSS = """
    Screen { background: $surface; }
    Header { background: $primary; color: $text; }
    Footer { background: $panel; }
    #main-container { height: 100%; }
    #left-panel { width: 50%; height: 100%; border: solid $primary; padding: 1; }
    #right-panel { width: 50%; height: 100%; border: solid $accent; padding: 1; }
    ListView { height: 1fr; border: solid $accent; background: $surface; }
    ListView > ListItem { padding: 0 1; }
    EnvValuePreview { height: 1fr; border: solid $primary; background: $surface; padding: 1; overflow-y: auto; }
    StatusBar { height: 3; border: solid $success; background: $surface; padding: 1; }
    Label { padding: 0 1; height: auto; }
    """

    BINDINGS = [
        Binding("q", "quit", "Quit", show=True),
        Binding("r", "refresh", "Refresh", show=True),
        Binding("c", "copy_entry", "Copy", show=True),
    ]

    def __init__(self) -> None:
        super().__init__()
        self._env_pairs: list[tuple[str, str]] = []
        self._env_lookup: dict[str, str] = {}
        self._selected_key: str = ""

    def compose(self) -> ComposeResult:
        platform_name = platform.system()
        yield Header(show_clock=True)
        with Horizontal(id="main-container"):
            with Vertical(id="left-panel"):
                yield Label(f"🌐 Environment Variables ({platform_name})")
                yield ListView(id="env-list")
            with Vertical(id="right-panel"):
                yield EnvValuePreview(id="preview")
                yield StatusBar(id="status")
        yield Footer()

    def on_mount(self) -> None:
        self.title = "Environment Explorer"
        self.sub_title = f"Platform: {platform.system()}"
        self._reload_environment()
        self._status().show_message("Ready. Select a variable to preview its value.", "info")

    def _reload_environment(self) -> None:
        self._env_pairs = collect_environment(os.environ)
        self._env_lookup = dict(self._env_pairs)
        self._populate_list()

    def _populate_list(self) -> None:
        list_view = self.query_one("#env-list", ListView)
        list_view.clear()
        for env_key, env_value in self._env_pairs:
            summary = format_summary(env_key, env_value, SUMMARY_LIMIT)
            list_view.append(EnvListItem(env_key, summary))
        self._status().show_message(f"Loaded {len(self._env_pairs)} environment variables.", "success")

    def _status(self) -> StatusBar:
        return self.query_one("#status", StatusBar)

    def _preview(self) -> EnvValuePreview:
        return self.query_one("#preview", EnvValuePreview)

    @on(ListView.Highlighted)
    def handle_highlight(self, event: ListView.Highlighted) -> None:
        if not isinstance(event.item, EnvListItem):
            return
        env_key = event.item.env_key()
        env_value = self._env_lookup.get(env_key, "")
        self._preview().show_value(env_key, env_value)
        self._status().show_message(f"Previewing {env_key}", "info")

    @on(ListView.Selected)
    def handle_selection(self, event: ListView.Selected) -> None:
        if not isinstance(event.item, EnvListItem):
            return
        env_key = event.item.env_key()
        self._selected_key = env_key
        env_value = self._env_lookup.get(env_key, "")
        self._preview().show_value(env_key, env_value)
        self._status().show_message(f"Selected {env_key}", "success")

    def action_refresh(self) -> None:
        self._reload_environment()
        self._status().show_message("Environment reloaded.", "success")

    def action_copy_entry(self) -> None:
        if self._selected_key == "":
            self._status().show_message("No variable selected.", "warning")
            return
        env_value = self._env_lookup.get(self._selected_key, "")
        payload = f"{self._selected_key}={env_value}"
        try:
            import pyperclip  # type: ignore[import]

            pyperclip.copy(payload)
            self._status().show_message(f"Copied {self._selected_key} to clipboard.", "success")
        except ImportError:
            self._status().show_message("pyperclip unavailable. Install it for clipboard support.", "warning")


def main() -> None:
    app = EnvExplorerApp()
    app.run()


if __name__ == "__main__":
    main()



"""Sessions management commands - lazy loading subcommands."""

from typing import Optional, Literal, Annotated
import typer


def balance_load(
    layout_path: Annotated[str, typer.Argument(..., help="Path to the layout.json file")],
    max_thresh: Annotated[int, typer.Option(..., "--max-threshold", "-m", help="Maximum tabs per layout")],
    thresh_type: Annotated[Literal["number", "n", "weight", "w"], typer.Option(..., "--threshold-type", "-t", help="Threshold type")],
    breaking_method: Annotated[Literal["moreLayouts", "ml", "combineTabs", "ct"], typer.Option(..., "--breaking-method", "-b", help="Breaking method")],
    output_path: Annotated[Optional[str], typer.Option(..., "--output-path", "-o", help="Path to write the adjusted layout.json file")] = None,
) -> None:
    """Adjust layout file to limit max tabs per layout, etc."""
    from machineconfig.scripts.python.helpers.helpers_sessions.sessions_impl import balance_load as impl
    impl(layout_path=layout_path, max_thresh=max_thresh, thresh_type=thresh_type, breaking_method=breaking_method, output_path=output_path)


def run(
    ctx: typer.Context,
    layout_path: Annotated[Optional[str], typer.Argument(..., help="Path to the layout.json file")] = None,
    max_tabs: Annotated[int, typer.Option(..., "--max-tabs", "-mt", help="A Sanity checker that throws an error if any layout exceeds the maximum number of tabs to launch.")] = 10,
    max_layouts: Annotated[int, typer.Option(..., "--max-layouts", "-ml", help="A Sanity checker that throws an error if the total number of *parallel layouts exceeds this number.")] = 10,
    sleep_inbetween: Annotated[float, typer.Option(..., "--sleep-inbetween", "-si", help="Sleep time in seconds between launching layouts")] = 1.0,
    monitor: Annotated[bool, typer.Option(..., "--monitor", "-m", help="Monitor the layout sessions for completion")] = False,
    parallel: Annotated[bool, typer.Option(..., "--parallel", "-p", help="Launch multiple layouts in parallel")] = False,
    kill_upon_completion: Annotated[bool, typer.Option(..., "--kill-upon-completion", "-k", help="Kill session(s) upon completion (only relevant if monitor flag is set)")] = False,
    choose: Annotated[Optional[str], typer.Option(..., "--choose", "-c", help="Comma separated names of layouts to be selected from the layout file passed")] = None,
    choose_interactively: Annotated[bool, typer.Option(..., "--choose-interactively", "-i", help="Select layouts interactively")] = False,
    subsitute_home: Annotated[bool, typer.Option(..., "--substitute-home", "-sh", help="Substitute ~ and $HOME in layout file with actual home directory path")] = False,
) -> None:
    """Launch terminal sessions based on a layout configuration file."""
    if layout_path is None:
        typer.echo(ctx.get_help())
        raise typer.Exit()
    from machineconfig.scripts.python.helpers.helpers_sessions.sessions_impl import run_layouts
    try:
        run_layouts(layout_path=layout_path, max_tabs=max_tabs, max_layouts=max_layouts, sleep_inbetween=sleep_inbetween, monitor=monitor, parallel=parallel, kill_upon_completion=kill_upon_completion, choose=choose, choose_interactively=choose_interactively, subsitute_home=subsitute_home)
    except ValueError as e:
        typer.echo(str(e))
        raise typer.Exit(1) from e


def create_template(
    name: Annotated[Optional[str], typer.Argument(..., help="Name of the layout template to create")] = None,
    num_tabs: Annotated[int, typer.Option(..., "--num-tabs", "-t", help="Number of tabs to include in the template")] = 3,
) -> None:
    """Create a layout template file."""
    from machineconfig.scripts.python.helpers.helpers_sessions.sessions_impl import create_template as impl
    impl(name=name, num_tabs=num_tabs)


def create_from_function(
    num_process: Annotated[int, typer.Option(..., "--num-process", "-n", help="Number of parallel processes to run")],
    path: Annotated[str, typer.Option(..., "--path", "-p", help="Path to a Python or Shell script file or a directory containing such files")] = ".",
    function: Annotated[Optional[str], typer.Option(..., "--function", "-f", help="Function to run from the Python file. If not provided, you will be prompted to choose.")] = None,
) -> None:
    """Create a layout from a function to run in multiple processes."""
    from machineconfig.scripts.python.helpers.helpers_sessions.sessions_multiprocess import create_from_function as impl
    impl(num_process=num_process, path=path, function=function)


def get_app() -> typer.Typer:
    layouts_app = typer.Typer(help="Layouts management subcommands", no_args_is_help=True, add_help_option=True, add_completion=False)

    layouts_app.command("create-from-function", no_args_is_help=True, short_help="[c] Create a layout from a function")(create_from_function)
    layouts_app.command("c", no_args_is_help=True, hidden=True)(create_from_function)

    layouts_app.command("run", no_args_is_help=True, help=run.__doc__, short_help="[r] Run the selected layout(s)")(run)
    layouts_app.command("r", no_args_is_help=True, help=run.__doc__, hidden=True)(run)

    layouts_app.command("balance-load", no_args_is_help=True, help=balance_load.__doc__, short_help="[b] Balance the load across sessions")(balance_load)
    layouts_app.command("b", no_args_is_help=True, help=balance_load.__doc__, hidden=True)(balance_load)

    layouts_app.command("create-template", no_args_is_help=False, help=create_template.__doc__, short_help="[t] Create a layout template file")(create_template)
    layouts_app.command("t", no_args_is_help=False, help=create_template.__doc__, hidden=True)(create_template)
    return layouts_app


def main() -> None:
    layouts_app = get_app()
    layouts_app()


if __name__ == "__main__":
    pass
